#!/usr/bin/python3.7
########################################################################################
# pipe.py - The pipe module for the matrix component.
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2021
########################################################################################

"""
The pipe module for the matrix component.

This module computes and returns the equation(s) associated with the pipe layer of the
PV-T collector for the matrix component.

The model works by arranging the system of differential equations as a matrix equation
such that the temperatures at each time step can be computed based on the coefficients
of the matrix which are computed based on the temperatures of the various components at
the previous time step, as well as various Physical and fundamental constants.

"""


import logging

from typing import List, Optional, Tuple, Union

import numpy

from .. import index_handler
from ..pvt_collector import pvt

from ...__utils__ import (
    OperatingMode,
    TemperatureName,
)
from ..__utils__ import WeatherConditions
from ..pvt_collector.element import Element
from ..pvt_collector.physics_utils import insulation_thermal_resistance

__all__ = "calculate_pipe_equation"


def calculate_pipe_equation(
    absorber_to_pipe_conduction: float,
    best_guess_temperature_vector: Union[List[float], numpy.ndarray],
    logger: logging.Logger,
    number_of_pipes: int,
    number_of_temperatures: int,
    number_of_x_elements: int,
    operating_mode: OperatingMode,
    pipe_to_htf_heat_transfer: float,
    previous_temperature_vector: Optional[numpy.ndarray],
    pvt_collector: pvt.PVT,
    resolution: Optional[int],
    element: Element,
    weather_conditions: WeatherConditions,
) -> Tuple[List[float], float]:
    """
    Returns a matrix row and resultant vector value representing the pipe equation.

    :param number_of_temperatures:
        The number of temperatures being modelled in the system.

    :return:
        A `tuple` containing:
        - the equation represented as a row in the matrix,
        - and the corresponding value in the resultant method.

    """

    # Compute the row equation
    row_equation: List[float] = [0] * number_of_temperatures

    if operating_mode.dynamic:
        pipe_internal_heat_change: float = (
            numpy.pi
            * (
                (pvt_collector.absorber.outer_pipe_diameter / 2) ** 2  # [m^2]
                - (pvt_collector.absorber.inner_pipe_diameter / 2) ** 2  # [m^2]
            )
            * element.length  # [m]
            * pvt_collector.absorber.pipe_density  # [kg/m^3]
            * pvt_collector.absorber.heat_capacity  # [J/kg*K]
            / resolution  # type: ignore  # [s]
        )
    else:
        pipe_internal_heat_change = 0

    pipe_to_surroundings_losses = (
        numpy.pi
        * (pvt_collector.absorber.outer_pipe_diameter / 2)  # [m]
        * element.length  # [m]
        / insulation_thermal_resistance(
            best_guess_temperature_vector,
            pvt_collector,
            index_handler.index_from_pipe_coordinates(
                number_of_pipes,
                number_of_x_elements,
                element.pipe_index,  # type: ignore
                pvt_collector,
                TemperatureName.pipe,
                element.y_index,
            ),
            weather_conditions,
        )  # [K*m^2/W]
    )

    # Compute the T_P(#, j) term.
    row_equation[
        index_handler.index_from_pipe_coordinates(
            number_of_pipes,
            number_of_x_elements,
            element.pipe_index,  # type: ignore
            pvt_collector,
            TemperatureName.pipe,
            element.y_index,
        )
    ] = (
        pipe_internal_heat_change  # [W/K]
        + absorber_to_pipe_conduction  # [W/K]
        + pipe_to_htf_heat_transfer  # [W/K]
        + pipe_to_surroundings_losses
    )

    # Compute the T_A(i, j) term.
    row_equation[
        index_handler.index_from_element_coordinates(
            number_of_x_elements,
            pvt_collector,
            TemperatureName.absorber,
            element.x_index,
            element.y_index,
        )
    ] = -1 * (absorber_to_pipe_conduction)

    # Compute the T_f(#, j) term.
    row_equation[
        index_handler.index_from_pipe_coordinates(
            number_of_pipes,
            number_of_x_elements,
            element.pipe_index,  # type: ignore
            pvt_collector,
            TemperatureName.htf,
            element.y_index,
        )
    ] = -1 * (pipe_to_htf_heat_transfer)

    # Compute the resultant vector value.
    resultant_vector_value = (
        # Ambient heat loss.
        pipe_to_surroundings_losses
        * weather_conditions.ambient_temperature  # [W]
    )

    if operating_mode.dynamic:
        resultant_vector_value += (
            # Internal heat change.
            pipe_internal_heat_change
            * previous_temperature_vector[  # type: ignore
                index_handler.index_from_pipe_coordinates(
                    number_of_pipes,
                    number_of_x_elements,
                    element.pipe_index,  # type: ignore
                    pvt_collector,
                    TemperatureName.pipe,
                    element.y_index,
                )
            ]
        )

    logger.debug(
        "Rough Pipe Temperature estimate: %s K.",
        int(
            resultant_vector_value
            / (pipe_to_surroundings_losses + pipe_internal_heat_change)
        ),
    )

    return row_equation, resultant_vector_value



#!/usr/bin/python3.7
########################################################################################
# __utils__.py - The utility module for this, my first, PV-T model! :O
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2020
########################################################################################

"""
The utility module for the PV-T model.

This module contains common functionality, strucutres, and types, to be used by the
various modules throughout the PVT model.

"""

import dataclasses
import enum
import logging
import os

import json

from dataclasses import dataclass
from typing import Any, Dict, Optional, Tuple, Union

import yaml

__all__ = (
    "BColours",
    "CarbonEmissions",
    "COARSE_RUN_RESOLUTION",
    "fourier_number",
    "get_logger",
    "INITIAL_CONDITION_PRECISION",
    "InvalidParametersError",
    "LOGGER_NAME",
    "OperatingMode",
    "ProgrammerJudgementFault",
    "MissingParametersError",
    "read_yaml",
    "save_data",
    "SteadyStateRun",
    "SystemData",
    "TotalPowerData",
)

# Coarse-run resolution:
#   The resolution in seconds for determining the initial conditions.
COARSE_RUN_RESOLUTION: int = 1800

# Initial-condition precision:
#   The prevision to reach when searching for consistent initial temperatures.
INITIAL_CONDITION_PRECISION: float = 1

# Logger directory:
#   The directory for storing the logs.
LOGGER_DIRECTORY = "logs"

# Logger name:
#   The name used for the internal logger.
LOGGER_NAME = "pvt_model"


@dataclass
class BColours:
    """
    Contains various colours used for pretty-printing out to the command-line on stdout.

    .. attribute:: FAIL
        Used for a failure message.

    .. attributes:: OKTEAL, WARNING, OKBLUE, HEADER, OKCYAN, OKGREEN
        Various colours used.

    .. attribute:: ENDC
        Used to reset the colour of the terminal output.

    .. attribute:: BOLD, UNDERLINE
        Used to format the text.

    """

    FAIL = "\033[91m"
    OKTEAL = "\033[92m"
    WARNING = "\033[93m"
    OKBLUE = "\033[94m"
    HEADER = "\033[95m"
    OKCYAN = "\033[96m"
    OKGREEN = "\033[97m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"


@dataclass
class CarbonEmissions:
    """
    Contains information about the carbon emissions produced.

    .. attribute:: electrical_carbon_produced
        The amount of CO2 equivalent produced, measured in kilograms.

    .. attribute:: electrical_carbon_saved
        The amount of CO2 saved by using PV-T, measured in kilograms.

    .. attribute:: heating_carbon_produced
        The amount of CO2 equivalent produced, measured in kilograms.

    .. attribute:: heating_carbon_saved
        The amount of CO2 equivalent saved by using the PV-T, measured in kilograms.

    """

    electrical_carbon_produced: float
    electrical_carbon_saved: float
    heating_carbon_produced: float
    heating_carbon_saved: float


class FileType(enum.Enum):
    """
    Tells what type of file is being used for the data.

    .. attribute:: YAML
        A YAML file is being used.

    .. attribute:: JSON
        A JSON file is being used.

    """

    YAML = 0
    JSON = 1


def fourier_number(
    conductive_length_scale: float,
    conductivity: float,
    density: float,
    heat_capacity: float,
    time_scale: float,
) -> float:
    """
    Calculates the Fourier coefficients based off the parameters passed in.

    The Fourier number is computed by:
        fourier_number = (
            conductivity * time_scale
        ) / (
            conductivy_length_scale ^ 2 * density * heat_capacity
        )

    :param conductive_length_scale:
        The length scale over which conduction occurs, measured in meters.

    :param conductivity:
        The thermal conductivity of the material, measured in Watts per meter Kelvin.

    :param density:
        The density of the medium, measured in kilograms per meter cubed.

    :param heat_capacity:
        The heat capcity of the conductive medium, measured in Joules per kilogram
        Kelvin.

    :param time_scale:
        The time scale of the simulation being run, measured in seconds.

    :return:
        The Fourier number based on these values.

    """

    f_num: float = (conductivity * time_scale) / (  # [W/m*K] * [s]
        heat_capacity  # [J/kg*K]
        * density  # [kg/m^3]
        * conductive_length_scale ** 2  # [m]^2
    )

    return f_num


def get_logger(
    disable_logging: bool, logger_name: str, verbose: bool
) -> logging.Logger:
    """
    Set-up and return a logger.

    :param disable_logging:
        If specified, file logging will be disabled.

    :param logger_name:
        The name of the logger to instantiate.

    :param verbose:
        Whether the logging is verbose (DEBUG reported) or not (INFO only).

    :return:
        The logger for the component.

    """

    # Create a logger with the current component name.
    logger = logging.getLogger(logger_name)
    # Create the logging directory if it doesn't exist.
    if not os.path.isdir(LOGGER_DIRECTORY):
        os.mkdir(LOGGER_DIRECTORY)
    # Rename old log files.
    append_index = 1
    if os.path.exists(os.path.join(LOGGER_DIRECTORY, f"{logger_name}.log")):
        while os.path.exists(
            os.path.join(LOGGER_DIRECTORY, f"{logger_name}.log.{append_index}")
        ):
            append_index += 1
    ch = logging.StreamHandler()
    if verbose:
        logger.setLevel(logging.DEBUG)
        ch.setLevel(logging.WARN)
    else:
        logger.setLevel(logging.INFO)
        ch.setLevel(logging.ERROR)
    # Create a formatter and add it to the handlers.
    formatter = logging.Formatter(
        "%(asctime)s: %(name)s: %(levelname)s: %(message)s",
        datefmt="%d/%m/%Y %I:%M:%S %p",
    )
    ch.setFormatter(formatter)
    # add the handlers to the logger
    logger.addHandler(ch)

    # If logging is not disabled, add a file handler.
    if not disable_logging:
        # Create the file handler.
        fh = logging.FileHandler(
            os.path.join(LOGGER_DIRECTORY, f"{logger_name}.log.{append_index}")
        )

        # Set the file-handler logging level correctly,
        if verbose:
            fh.setLevel(logging.DEBUG)
        else:
            fh.setLevel(logging.INFO)

        # Add the file handler to the logger.
        fh.setFormatter(formatter)
        logger.addHandler(fh)

    return logger


class InvalidParametersError(Exception):
    """
    Raised when some parameters have been specified incorrectly.

    """

    def __init__(self, message: str, variable_name: str) -> None:
        """
        Instantiate an invalid parameters error.

        :param message:
            An appended message to display to the user.

        :param variable_name:
            The name of the variable for which insufficient parameters were specified.

        """

        super().__init__(
            f"Invalid parameters when determining '{variable_name}': {message}."
        )


class MissingParametersError(Exception):
    """
    Raised when not all parameters have been specified that are needed to instantiate.

    """

    def __init__(self, class_name: str, message: str) -> None:
        """
        Instantiate a missing parameters error.

        :param class_name:
            The class for which insufficient parameters were specified.

        :param message:
            An appended message to display to the user.

        """

        super().__init__(
            f"Missing parameters when initialising a '{class_name}' class: {message}."
        )


@dataclass
class OperatingMode:
    """
    Contains information about the mode of operation of the model.

    .. attribute:: coupled
        Whether the system is coupled (True) or decoupled (False)

    .. attribute:: dynamic
        Whether the system is dyanmic (True) or steady-state (False).

    """

    coupled: bool
    dynamic: bool

    @property
    def decoupled(self) -> bool:
        """
        Returns whether the operating mode is decoupled.

        :return:
            A `bool` giving whether the system is coupled (False) or decoupled (True).

        """

        return not self.coupled

    @property
    def steady_state(self) -> bool:
        """
        Returns whether the operating mode is steady-state.

        :return:
            A `bool` giving whether the system is dynamic (False) or steady-state
            (True).

        """

        return not self.dynamic


class ProgrammerJudgementFault(Exception):
    """
    Raised when an error is hit due to poor programming.

    """

    def __init__(self, message: str) -> None:
        """
        Instantiate a programmer judgement fault error.

        :param message:
            A message to append when displaying the error to the user.

        """

        super().__init__(f"A programmer judgement fault has occurred: {message}")


def read_yaml(yaml_file_path: str) -> Dict[Any, Any]:
    """
    Read in some yaml data and return it.

    :param yaml_file_path:
        The path to the yaml data to read in.

    :return:
        A `dict` containing the data read in from the yaml file.

    """

    logger = logging.getLogger(LOGGER_NAME)

    # Open the yaml data and read it.
    if not os.path.isfile(yaml_file_path):
        logger.error(
            "A YAML data file, '%s', could not be found. Exiting...", yaml_file_path
        )
        raise FileNotFoundError(yaml_file_path)
    with open(yaml_file_path) as f:
        try:
            data: Dict[Any, Any] = yaml.safe_load(f)
        except yaml.parser.ParserError as e:
            logger.error("Failed to read YAML file '%s'.", yaml_file_path)
            print(f"Failed to parse YAML. Internal error: {str(e)}")
            raise

    logger.info("Data successfully read from '%s'.", yaml_file_path)
    return data


@dataclass
class SystemData:
    """
    Contains information about the system at a given time step.

    .. attribute:: ambient_temperature
        The ambient temperature, measured in Celcius.

    .. attribute:: bulk_water_temperature
        The temperature of the bulk water, measured in Celcius.

    .. attribute:: absorber_temperature
        The temperature of the absorber layer, measured in Celcius.

    .. attribute:: date
        A `str` giving the current date.

    .. attribute:: electrical_efficiency
        The electrical efficiency of the PV layer of the collector.

    .. attribute:: exchanger_temperature_drop
        The temperature drop through the heat exchanger in the tank, measured in Kelvin
        or Celcius. (As it is a temperature difference, the two scales are equivalent.)

    .. attribute:: glass_temperature
        The temperature of the glass layer, measured in Celcius.  Can be `None` if no
        glass present.

    .. attribute:: pipe_temperature
        The temperature of the pipe, measured in Celcius.

    .. attribute:: pv_temperature
        The temperature of the PV layer, measured in Celcius. Can be `None` if no PV
        present.

    .. attribute:: reduced_temperature
        The reduced temperature of the PV-T absorber, measured in Celcius.

    .. attribute:: sky_temperature
        The temperature of the sky, measured in Celcius.

    .. attribute:: solar_irradiance
        The solar irradiance incident on the PV-T collector, measured in Watts per meter
        squared.

    .. attribute:: tank_temperature
        The temperature of the hot-water tank, measured in Celcius.

    .. attribute:: thermal_efficiency
        The thermal efficiency of the system.

    .. attribute:: upper_glass_temperature
        The temperature of the upper-glass (i.e., double-glazing) temperature, measured
        in Celcius. Can be `None` if no double-glazing present.

    .. attribute:: collector_input_temperature
        The temperature of the HTF inputted into the absorber, measured in Celcius.
        This can be set to `None` if no data is recorded.

    .. attribute:: collector_output_temperature
        The temperature of the HTF outputted from the absorber, measured in Celcius.
        This can be set to `None` if no data is recorded.

    .. attribute:: collector_temperature_gain
        The temperature gain of the HTF through the thermal collector.

    .. attribute:: layer_temperature_map_bulk_water
        A mapping between coordinate and temperature for the bulk water within the
        pipes.

    .. attribute:: layer_temperature_map_absorber
        A mapping between coordinate and temperature for elements within the absorber
        layer.

    .. attribute:: layer_temperature_map_glass
        A mapping between coordinate and temperature for elements within the glass layer.

    .. attribute:: layer_temperature_map_pipe
        A mapping between coordinate and temperature for the pipes.

    .. attribute:: layer_temperature_map_pv
        A mapping between coordinate and temperature for elements within the pv layer.

    .. attribute:: layer_temperature_map_upper_glass
        A mapping between coordinate and temperature for elements within the upper-glass
        (i.e., double-glazing) layer.

    .. attribute:: reduced_collector_temperature
        The reduced temperature of the collector.

    .. attribute:: time
        A `str` giving the current time, can be set to `None` for steady-state runs.

    """

    absorber_temperature: float
    ambient_temperature: float
    bulk_water_temperature: float
    date: str
    electrical_efficiency: Optional[float]
    glass_temperature: Optional[float]
    exchanger_temperature_drop: Optional[float]
    pipe_temperature: float
    pv_temperature: float
    reduced_collector_temperature: Optional[float]
    sky_temperature: float
    solar_irradiance: float
    tank_temperature: Optional[float]
    thermal_efficiency: Optional[float]
    upper_glass_temperature: Optional[float]
    collector_input_temperature: Optional[float] = None
    collector_output_temperature: Optional[float] = None
    collector_temperature_gain: Optional[float] = None
    layer_temperature_map_bulk_water: Optional[Dict[str, float]] = None
    layer_temperature_map_absorber: Optional[Dict[str, float]] = None
    layer_temperature_map_glass: Optional[Dict[str, float]] = None
    layer_temperature_map_pipe: Optional[Dict[str, float]] = None
    layer_temperature_map_pv: Optional[Dict[str, float]] = None
    layer_temperature_map_upper_glass: Optional[Dict[str, float]] = None
    reduced_temperature: Optional[float] = None
    time: Optional[str] = None


class TemperatureName(enum.Enum):
    """
    Used for keeping track of the temperature value being used.

    """

    upper_glass = 0
    glass = 1
    pv = 2
    absorber = 3
    pipe = 4
    htf = 5
    htf_in = 6
    htf_out = 7
    collector_in = 8
    collector_out = 9
    tank_in = 10
    tank_out = 11
    tank = 12


@dataclass
class TotalPowerData:
    """
    Contains information about the total power generated by the system.

    .. attribute:: electricity_supplied
        The electricity supplied, measured in Joules.

    .. attribute:: electricity_demand
        The electricity demand, measured in Joules.

    .. attribute:: heating_supplied
        The heatimg supplied, measured in Joules.

    .. attribute:: heating_demand
        The heating demand, measured in Joules.

    """

    electricity_supplied: float = 0
    electricity_demand: float = 0
    heating_supplied: float = 0
    heating_demand: float = 0

    def increment(
        self,
        electricity_supplied_incriment,
        electricity_demand_incriment,
        heating_supplied_increment,
        heating_demand_increment,
    ) -> None:
        """
        Updates the values by incrementing with those supplied.

        :param electricity_supplied_increment:
            Electricity supplied to add, measured in Joules.
        :param electricity_demand_increment:
            Electricity demand to add, measured in Joules.
        :param heating_supplied_increment:
            Heating supplied to add, measured in Joules.
        :param heating_demand_increment:
            Heating demand to add, measured in Joules.

        """

        self.electricity_supplied += electricity_supplied_incriment
        self.electricity_demand += electricity_demand_incriment
        self.heating_supplied += heating_supplied_increment
        self.heating_demand += heating_demand_increment


def save_data(
    file_type: FileType,
    logger: logging.Logger,
    operating_mode: OperatingMode,
    output_file_name: str,
    system_data: Dict[float, SystemData],
    carbon_emissions: Optional[CarbonEmissions] = None,
    total_power_data: Optional[TotalPowerData] = None,
) -> None:
    """
    Save data when called. The data entry should be appended to the file.

    :param file_type:
        The file type that's being saved.

    :param logger:
        The logger used for the run.

    :param operating_mode:
        The operating mode for the run.

    :param output_file_name:
        The destination file name.

    :param system_data:
        The data to save.

    :param carbon_emissions:
        The carbon emissions data for the run.

    :param total_power_data:
        The total power data for the run.

    """

    # Convert the system data entry to JSON-readable format
    system_data_dict: Dict[Union[float, str], Union[str, Dict[str, Any]]] = {
        key: dataclasses.asdict(value) for key, value in system_data.items()
    }
    logger.info(
        "Saving data: System data successfully converted to json-readable format."
    )

    # If we're saving YAML data part-way through, then append to the file.
    if file_type == FileType.YAML:
        logger.info("Saving as YAML format.")
        with open(f"{output_file_name}.yaml", "a") as output_yaml_file:
            yaml.dump(
                system_data_dict,
                output_yaml_file,
            )

    # If we're dumping JSON, open the file, and append to it.
    if file_type == FileType.JSON:
        logger.info("Saving as JSON format.")
        # Append the total power and emissions data for the run.
        if total_power_data is not None:
            system_data_dict.update(dataclasses.asdict(total_power_data))  # type: ignore
            logger.info("Total power data updated.")
        if carbon_emissions is not None:
            system_data_dict.update(dataclasses.asdict(carbon_emissions))  # type: ignore
            logger.info("Carbon emissions updated.")

        # Append the data type for the run.
        system_data_dict["data_type"] = "{coupling}_{timing}".format(
            coupling="coupled" if operating_mode.coupled else "decoupled",
            timing="steady_state" if operating_mode.steady_state else "dynamic",
        )
        logger.info("Data type added successfully.")

        # Save the data
        # If this is the initial dump, then create the file.
        if not os.path.isfile(f"{output_file_name}.json"):
            logger.info("Attempting first dump to a non-existent file.")
            logger.info("Output file name: %s", output_file_name)
            with open(f"{output_file_name}.json", "w") as output_json_file:
                json.dump(
                    system_data_dict,
                    output_json_file,
                    indent=4,
                )
        else:
            with open(f"{output_file_name}.json", "r+") as output_json_file:
                # Read the data and append the current update.
                filedata = json.load(output_json_file)
                filedata.update(system_data_dict)
                # Overwrite the file with the updated data.
                output_json_file.seek(0)
                json.dump(
                    filedata,
                    output_json_file,
                    indent=4,
                )


@dataclass(frozen=True)
class SteadyStateRun:
    """
    Represents a steady-state run which is to be carreid out.

    .. attribute:: ambient_temperature
        The ambient temperature, in Celcius, to use for the run.

    .. attribute:: collector_input_temperature
        The collector input temperature, in Celcius, to use for the run.

    .. attribute:: irradiance
        The incidient irradiance, in Watts per meter squared, to use for the run.

    .. attribute:: mass_flow_rate
        The mass-flow-rate, in litres per hour, to use for the run.

    .. attribute:: wind_speed
        The wind speed, in meters per second, to use for the run.

    """

    ambient_temperature: float
    collector_input_temperature: float
    irradiance: float
    mass_flow_rate: float
    wind_speed: float

    @classmethod
    def from_data(cls, data: Dict[str, float]) -> Any:
        """
        Generates a :class:`SteadyStateData` instance based on the input data.

        :param data:
            The input data, extracted from the state-state inputs file.

        """

        return cls(
            data["ambient_temperature"],
            data["collector_input_temperature"],
            data["irradiance"],
            data["mass_flow_rate"],
            data["wind_speed"],
        )

    def to_tuple(self) -> Tuple[float, float, float, float, float]:
        """
        Returns a `tuple` containing the data.

        :returns:
            A `tuple` containing the data.

        """

        return (
            self.ambient_temperature,
            self.collector_input_temperature,
            self.irradiance,
            self.mass_flow_rate,
            self.wind_speed,
        )



#!/usr/bin/python3.7
########################################################################################
# coupled.py - Module for executing coupled system runs.
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2021
########################################################################################

"""
The coupled-run module for the PV-T model.

This module exposes methods needed to carry out a coupled run of the system.

A coupled run treates the PV-T collector as connected to the rest of the system (tank
etc.) with demand profiles being utilised accordingly.

"""

import logging

from typing import Dict, List, Optional, Tuple, Union

import numpy

from dateutil.relativedelta import relativedelta

from .. import convergent_solver, exchanger, load, tank, weather
from ..pvt_collector import pvt

from ...__utils__ import BColours, OperatingMode, ProgrammerJudgementFault, SystemData
from ..__utils__ import DivergentSolutionError, time_iterator
from ..constants import DEFAULT_INITIAL_DATE_AND_TIME
from .__utils__ import system_data_from_run

__all__ = ("coupled_dynamic_run",)


def coupled_dynamic_run(
    cloud_efficacy_factor: float,
    days: Optional[int],
    heat_exchanger: exchanger.Exchanger,
    hot_water_tank: tank.Tank,
    initial_month: int,
    initial_system_temperature_vector: List[float],
    load_system: load.LoadSystem,
    logger: logging.Logger,
    minutes: Optional[int],
    months: Optional[int],
    number_of_pipes: int,
    number_of_temperatures: int,
    number_of_x_elements: int,
    number_of_y_elements: int,
    operating_mode: OperatingMode,
    pvt_collector: pvt.PVT,
    resolution: int,
    save_2d_output: bool,
    start_time: int,
    weather_forecaster: weather.WeatherForecaster,
) -> Tuple[numpy.ndarray, Dict[float, SystemData]]:
    """
    Carries out a dynamic run of the system.

    :param cloud_efficacy_factor:
        How effective the cloud cover is, from 0 (no effect) to 1 (no solar irradiance
        makes it through the cloud layer).

    :param days:
        How many days the system should run the simulation for.

    :param heat_exchanger:
        The heat exchanger between the HTF and the hot-water tank.

    :param hot_water_tank:
        A :class:`tank.Tank` instance representing the hot-water tank being included in
        the system.

    :param initial_month:
        The month for which the simulation should start to be run.

    :param initial_system_temperature_vector:
        The set of system temperatures, expressed as a `list`, at the beginning of the
        run.

    :param load_system:
        The load system for the run.

    :param logger:
        The logger to be used for the run.

    :param minutes:
        The number of minutes for which to run the simulation, if specified. Otherwise,
        `None`.

    :param months:
        The number of months for which to run the simulation, if specified. Otherwise,
        `None`.

    :param number_of_pipes:
        The number of pipes attached to the absorber.

    :param number_of_temperatures:
        The number of temperatures being modelled.

    :param number_of_x_elements:
        The number of elements in the x direction.

    :param number_of_y_elements:
        The number of elements in the y direction.

    :param operating_mode:
        The operating mode of the system.

    :param pvt_collector:
        A :class:`pvt.PVT` instance representing the PVT panel being modelled.

    :param resolution:
        The temporal resolution (in seconds) for the run.

    :param save_2d_output:
        Whether the 2D output should be saved (True) or not (False).

    :param start_time:
        The start time, expressed in hours from midnight, at which to begin running the
        model.

    :param weather_forecaster:
        The weather forecaster for the system.

    :return:
        A `tuple` containing:
        - the final system-state reached by the model,
        - the system data for the run.

    """

    # Set up a holder for information about the system.
    system_data: Dict[float, SystemData] = dict()

    # Set up the time iterator.
    if minutes is None and days is None and months is None:
        raise ProgrammerJudgementFault(
            "{}Either minutes, days or months must be specified for dynamic ".format(
                BColours.FAIL
            )
            + "runs.{}".format(BColours.ENDC)
        )

    num_months = (
        (initial_month if initial_month is not None else 1)
        - 1
        + (months if months is not None else 0)
    )  # [months]
    start_month = (
        initial_month
        if 1 <= initial_month <= 12
        else DEFAULT_INITIAL_DATE_AND_TIME.month
    )  # [months]
    initial_date_and_time = DEFAULT_INITIAL_DATE_AND_TIME.replace(
        hour=start_time, month=start_month
    )
    if days is None:
        final_date_and_time = initial_date_and_time + relativedelta(
            months=num_months % 12, years=num_months // 12
        )
    else:
        final_date_and_time = initial_date_and_time + relativedelta(days=days)

    logger.info(
        "Beginning itterative model:\n  Running from: %s\n  Running to: %s",
        str(initial_date_and_time),
        str(final_date_and_time),
    )

    time_iterator_step = relativedelta(seconds=resolution)

    weather_conditions = weather_forecaster.get_weather(
        pvt_collector.latitude,
        pvt_collector.longitude,
        cloud_efficacy_factor,
        initial_date_and_time,
    )

    system_data[0] = system_data_from_run(
        initial_date_and_time.date(),
        initial_date_and_time,
        number_of_pipes,
        number_of_x_elements,
        operating_mode,
        pvt_collector,
        save_2d_output,
        initial_system_temperature_vector,
        initial_date_and_time.time(),
        weather_conditions,
    )

    previous_run_temperature_vector: Union[
        List[float], numpy.ndarray
    ] = initial_system_temperature_vector

    for run_number, date_and_time in enumerate(
        time_iterator(
            first_time=initial_date_and_time,
            last_time=final_date_and_time,
            resolution=resolution,
            timezone=pvt_collector.timezone,
        )
    ):

        logger.debug(
            "Time: %s: Beginning internal run. Previous temperature vector: T=%s",
            date_and_time.strftime("%d/%m/%Y %H:%M:%S"),
            previous_run_temperature_vector,
        )

        # Determine the "i+1" time.
        next_date_and_time = date_and_time + time_iterator_step

        # Determine the "i+1" current hot-water load.
        current_hot_water_load = (
            load_system[
                (load.ProfileType.HOT_WATER, next_date_and_time)
            ]  # [litres/hour]
            / 3600  # [seconds/hour]
        )  # [kg/s]

        # Determine the "i+1" current weather conditions.
        try:
            weather_conditions = weather_forecaster.get_weather(
                pvt_collector.latitude,
                pvt_collector.longitude,
                cloud_efficacy_factor,
                next_date_and_time,
            )
        except KeyError as e:
            logger.error("Failed to get weather conditions, do all days have profiles?")
            raise

        try:
            current_run_temperature_vector = (
                convergent_solver.solve_temperature_vector_convergence_method(
                    current_hot_water_load=current_hot_water_load,
                    heat_exchanger=heat_exchanger,
                    hot_water_tank=hot_water_tank,
                    logger=logger,
                    next_date_and_time=next_date_and_time,
                    number_of_pipes=number_of_pipes,
                    number_of_temperatures=number_of_temperatures,
                    number_of_x_elements=number_of_x_elements,
                    number_of_y_elements=number_of_y_elements,
                    operating_mode=operating_mode,
                    previous_run_temperature_vector=previous_run_temperature_vector,
                    pvt_collector=pvt_collector,
                    resolution=resolution,
                    run_one_temperature_vector=previous_run_temperature_vector,
                    weather_conditions=weather_conditions,
                )
            )
        except DivergentSolutionError as e:
            logger.error(
                "A divergent solution was reached at %s:%s",
                date_and_time.strftime("%D/%M/%Y %H:%M:%S"),
                str(e),
            )
            raise

        # Save the system data output and 2D profiles.
        system_data[run_number + 1] = system_data_from_run(
            next_date_and_time.date(),
            initial_date_and_time,
            number_of_pipes,
            number_of_x_elements,
            operating_mode,
            pvt_collector,
            save_2d_output,
            current_run_temperature_vector,
            next_date_and_time.time(),
            weather_conditions,
        )
        previous_run_temperature_vector = current_run_temperature_vector
        logger.info(
            "Time: %s: Internal run complete with consistent conditions. Moving to the next time step.",
            date_and_time.strftime("%d/%m/%Y %H:%M:%S"),
        )

    return numpy.asarray(current_run_temperature_vector), system_data



#!/usr/bin/python3.7
########################################################################################
# tank.py - The tank module for this, my first, PV-T model! :O
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2020
########################################################################################

"""
The tank module for the PV-T model.

This module represents the hot-water tank.

"""

from .constants import HEAT_CAPACITY_OF_WATER

__all__ = (
    "net_enthalpy_gain",
    "Tank",
)


def net_enthalpy_gain(
    delivery_temperature: float,
    mains_water_temperature: float,
    water_demand_volume: float,
) -> float:
    """
    Computes the net enthalpy gain by the tank due to delivering water, in Watts.

    :param delivery_temperature:
        The delivery temperature of the hot-water tank, measured in Kelvin.

    :param mains_water_temperature:
        The temperature of the mains water used to replace any water removed from the
        hot-water tank, measured in Kelvin.

    :param water_demand_volume:
        The volume of water removed from the tank, measured in litres. This will be the
        same as the mass of water removed from the tank, measured in kilograms, due to
        the density of water.

    """

    return (
        water_demand_volume  # [kg]
        * HEAT_CAPACITY_OF_WATER  # [J/kg*K]
        * (mains_water_temperature - delivery_temperature)  # [K]
    )


class Tank:
    """
    Represents a hot-water tank.

    .. attribute:: area
        The surface area of the tank, measured in meters squared.

    .. attribute:: diameter
        The diameter of the tank, measured in meters.

    .. attribute:: heat_capacity
        The heat capacity of the water, measured in Joules per kilogram Kelvin.

    .. attribute:: heat_loss_coefficient
        The heat lost from the tank, measured in Watts per meter squared Kelvin.

    .. attribute:: mass
        The mass of water in the hot-water tank, measured in kilograms.

    """

    def __init__(
        self,
        area: float,
        diameter: float,
        heat_capacity: float,
        heat_loss_coefficient: float,
        mass: float,
    ) -> None:
        """
        Instantiate a hot-water tank.

        :param area:
            The surface area of the tank, measured in meters squared.

        :param diameter:
            The diameter of the hot-water tank being mnodelled, measured in meters.

        :param heat_capacity:
            The heat capacity of water within the tank, measured in Joules per kilogram
            Kelvin.

        :param heat_loss_coefficient:
            The heat lost from the tank, measured in Watts per meter squared Kelvin.

        :param mass:
            The mass of water that can be held within the tank, measured in kilograms.

        """

        self.area = area
        self.diameter = diameter
        self.heat_capacity = heat_capacity
        self.heat_loss_coefficient = heat_loss_coefficient
        self.mass = mass

    def __repr__(self) -> str:
        """
        Returns a nice representation of the hot-water tank.

        :return:
            A `str` giving a nice representation of the hot-water tank.

        """

        return (
            "Tank("
            f"area: {self.area}m^2, "
            f"diameter: {self.diameter}m, "
            f"heat_capacity: {self.heat_capacity}J/kg*K, "
            f"heat_loss_coefficient: {self.heat_loss_coefficient}W/m^2*K, "
            f"mass: {self.mass}kg, "
            ")"
        )

    def heat_loss(
        self, ambient_tank_temperature: float, tank_temperature: float
    ) -> float:
        """
        Computes the heat loss through the walls of the tank, measured in Watts.

        A value of 573 W/K is used by Maria's paper.

        :param ambient_temperature:
            The temperature of the air surrounding the hot-water tank, measured in
            Kelvin.

        :param tank_temperature:
            The temperature of the fluid within the tank, measured in Kelvin.

        :return:
            The heat loss from the tank, measured in Watts.

        """

        return (
            self.area  # [m^2]
            * self.heat_loss_coefficient  # [W/m^2*K]
            * (tank_temperature - ambient_tank_temperature)  # [K]
        )  # [W]



#!/usr/bin/python3.7
########################################################################################
# index_handler.py - The index solver module for the PVT model component.
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2021
########################################################################################

"""
The index module for the PVT model.

When solving the temperatures of the PVT model using a matrix method, it is necessary to
keep track of a mapping between the temperatures of the various components and a unique
index which can be used to represent the temperatures of the system as a single 1D
vector or `tuple`. The work for doing so is done here, in the index module.

"""

from typing import Optional, Tuple

from ..__utils__ import (
    ProgrammerJudgementFault,
    TemperatureName,
)
from .pvt_collector import pvt

__all__ = (
    "index_from_pipe_coordinates",
    "index_from_element_coordinates",
    "index_from_temperature_name",
    "num_temperatures",
    "temperature_name_from_index",
    "x_coordinate",
    "y_coordinate",
)


def _calculate_number_of_temperatures(
    pvt_collector: pvt.PVT,
) -> Tuple[int, int, int, int, int, int]:
    """
    Calculates, based off the PVT collector, the number of temperatures in each layer.

    :param pvt_collector:
        The PVT collector being modelled.

    :return:
        A `tuple` containing:
        - the number of upper-glass (i.e., double-glazing) temperatures in the
          collector,
        - the number of glass temperatures in the collector,
        - the number of pv temperatures in the collector,
        - the number of absorber temperatures in the collector,
        - the number of panel (i.e., glass, pv, and absorber) temperatures in the
          collector,
        - the number of pipe/htf temperatures in the collector,

    """

    num_upper_glass_temperatures = sum(
        [element.upper_glass for element in pvt_collector.elements.values()]
    )
    num_glass_temperatures = sum(
        [element.glass for element in pvt_collector.elements.values()]
    )
    num_pv_temperatures = sum(
        [element.pv for element in pvt_collector.elements.values()]
    )
    num_absorber_temperatures = sum(
        [element.absorber for element in pvt_collector.elements.values()]
    )
    num_panel_temperatures: int = (
        num_upper_glass_temperatures
        + num_glass_temperatures
        + num_pv_temperatures
        + num_absorber_temperatures
    )
    num_fluid_temperatures = sum(
        [element.pipe for element in pvt_collector.elements.values()]
    )

    return (
        num_upper_glass_temperatures,
        num_glass_temperatures,
        num_pv_temperatures,
        num_absorber_temperatures,
        num_panel_temperatures,
        num_fluid_temperatures,
    )


def _get_index(  # pylint: disable=too-many-branches
    temperature_name: TemperatureName,
    pvt_collector: pvt.PVT,
    *,
    number_of_pipes: Optional[int] = None,
    number_of_x_elements: Optional[int] = None,
    pipe_number: Optional[int] = None,
    x_coord: Optional[int] = None,
    y_coord: Optional[int] = None
) -> int:
    """
    Returns a unique index based on the inputs.

    :param temperature_name:
        The name of the temperature being indexed.

    :param pvt_collector:
        The :class:`pvt.PVT` instance representing the PVT collector being modelled.

    :param number_of_pipes:
        The number of HTF pipes in the absorber.

    :param number_of_x_elements:
        The number of elements in the x direction for the absorber model being run.

    :param pipe_number:
        The pipe number for which a coordinate should be returned.

    :param x_coord:
        The x coordinate of the element whose index is being determined.

    :param y_coord:
        The y coordinate of the element whose index is being determined.

    :return:
        A unique index describing the element.

    :raises: ProgrammerJudegementFault
        This is raised if the method for defining the index for the temperature passed
        in is not defined in this method.

    """

    index: Optional[int] = None

    # Fetch the number of temperatures of each type.
    (
        num_upper_glass_temperatures,
        num_glass_temperatures,
        num_pv_temperatures,
        _,
        num_panel_temperatures,
        num_fluid_temperatures,
    ) = _calculate_number_of_temperatures(pvt_collector)

    if temperature_name == TemperatureName.upper_glass:
        if number_of_x_elements is None or x_coord is None or y_coord is None:
            raise ProgrammerJudgementFault(
                "Not all parameters needed were passed in to uniquely determine an "
                "upper-glass (i.e., double-glazing) layer index_handler."
            )
        index = int(number_of_x_elements * y_coord + x_coord)
    if temperature_name == TemperatureName.glass:
        if number_of_x_elements is None or x_coord is None or y_coord is None:
            raise ProgrammerJudgementFault(
                "Not all parameters needed were passed in to uniquely determine a "
                "glass layer index_handler."
            )
        index = num_upper_glass_temperatures + int(
            number_of_x_elements * y_coord + x_coord
        )
    if temperature_name == TemperatureName.pv:
        if number_of_x_elements is None or x_coord is None or y_coord is None:
            raise ProgrammerJudgementFault(
                "Not all parameters needed were passed in to uniquely determine a "
                "pv layer index_handler."
            )
        index = (
            num_upper_glass_temperatures
            + num_glass_temperatures
            + int(number_of_x_elements * y_coord + x_coord)
        )
    if temperature_name == TemperatureName.absorber:
        if number_of_x_elements is None or x_coord is None or y_coord is None:
            raise ProgrammerJudgementFault(
                "Not all parameters needed were passed in to uniquely determine an "
                "absorber layer index_handler."
            )
        index = (
            num_upper_glass_temperatures
            + num_glass_temperatures
            + num_pv_temperatures
            + int(number_of_x_elements * y_coord + x_coord)
        )
    if temperature_name == TemperatureName.pipe:
        if number_of_pipes is None or pipe_number is None or y_coord is None:
            raise ProgrammerJudgementFault(
                "Not all parameters needed were passed in to uniquely determine a "
                "pipe index_handler."
            )
        index = num_panel_temperatures + int((number_of_pipes * y_coord + pipe_number))
    if temperature_name == TemperatureName.htf:
        if number_of_pipes is None or pipe_number is None or y_coord is None:
            raise ProgrammerJudgementFault(
                "Not all parameters needed were passed in to uniquely determine an "
                "htf index_handler."
            )
        index = (
            num_panel_temperatures
            + num_fluid_temperatures
            + int((number_of_pipes * y_coord + pipe_number))
        )
    if temperature_name == TemperatureName.htf_in:
        if number_of_pipes is None or pipe_number is None or y_coord is None:
            raise ProgrammerJudgementFault(
                "Not all parameters needed were passed in to uniquely determine an "
                "htf-input index_handler."
            )
        index = (
            num_panel_temperatures
            + 2 * num_fluid_temperatures
            + int((number_of_pipes * y_coord + pipe_number))
        )
    if temperature_name == TemperatureName.htf_out:
        if number_of_pipes is None or pipe_number is None or y_coord is None:
            raise ProgrammerJudgementFault(
                "Not all parameters needed were passed in to uniquely determine an "
                "htf-output index_handler."
            )
        index = (
            num_panel_temperatures
            + 3 * num_fluid_temperatures
            + int((number_of_pipes * y_coord + pipe_number))
        )
    if temperature_name == TemperatureName.collector_in:
        index = num_panel_temperatures + 4 * num_fluid_temperatures
    if temperature_name == TemperatureName.collector_out:
        index = num_panel_temperatures + 4 * num_fluid_temperatures + 1
    if temperature_name == TemperatureName.tank:
        index = num_panel_temperatures + 4 * num_fluid_temperatures + 2
    if temperature_name == TemperatureName.tank_in:
        index = num_panel_temperatures + 4 * num_fluid_temperatures + 3
    if temperature_name == TemperatureName.tank_out:
        index = num_panel_temperatures + 4 * num_fluid_temperatures + 4

    # Return the index if assigned, else, raise an error.
    if index is not None:
        return index
    raise ProgrammerJudgementFault(
        "An attempt was made to fetch an index using an undefined method."
    )


def index_from_pipe_coordinates(
    number_of_pipes: int,
    number_of_x_elements: int,
    pipe_number: int,
    pvt_collector: pvt.PVT,
    temperature_name: TemperatureName,
    y_coord: int,
) -> int:
    """
    Computes an index for a elemented pipe based on the coordinates of the element.

    :param number_of_pipes:
        The number of HTF pipes in the absorber.

    :param number_of_x_elements:
        The number of elements in the x direction along the panel.

    :param pipe_number:
        The number of the pipe.

    :param pvt_collector:
        The pvt collector being modelled.

    :param temperature_name:
        The name of the layer/temperature type being computed.

    :param y_coord:
        The y coordinate of the element.

    :return:
        The index describing teh pipe and element uniquely.

    """

    return _get_index(
        temperature_name,
        pvt_collector,
        number_of_pipes=number_of_pipes,
        number_of_x_elements=number_of_x_elements,
        pipe_number=pipe_number,
        y_coord=y_coord,
    )


def index_from_element_coordinates(
    number_of_x_elements: int,
    pvt_collector: pvt.PVT,
    temperature_name: TemperatureName,
    x_coord: int,
    y_coord: int,
) -> int:
    """
    Computes an index for a elemented layer based on the coordinates of the element.

    :param number_of_x_elements:
        The number of elements in the x direction along the panel.

    :param pvt_collector:
        The pvt collector being modelled.

    :param temperature_name:
        The name of the layer/temperature type being computed.

    :param x_coord:
        The x coordinate of the element.

    :param y_coord:
        The y coordinate of the element.

    :return:
        The index describing the layer and the element uniquely.

    """

    return _get_index(
        temperature_name,
        pvt_collector,
        number_of_x_elements=number_of_x_elements,
        x_coord=x_coord,
        y_coord=y_coord,
    )


def index_from_temperature_name(
    pvt_collector: pvt.PVT,
    temperature_name: TemperatureName,
) -> int:
    """
    Computes an index for a body/temperature based solely on the name.

    :param pvt_collector:
        The PVT collector being modelled.

    :param temperature_name:
        The name of the temperature/layer being computed.

    :return:
        An index uniquely describing the temperature name.

    """

    return _get_index(
        temperature_name,
        pvt_collector,
    )


def num_temperatures(pvt_collector: pvt.PVT) -> int:
    """
    Returns the number of temperature variables being modelled.

    :param pvt_collector:
        The pvt panel being modelled.

    :return:
        The total number of temperatures being modelled.

    """

    return (
        sum(
            [
                element.upper_glass
                + element.glass
                + element.pv
                + element.absorber
                + 4 * element.pipe
                for element in pvt_collector.elements.values()
            ]
        )
        + 5
    )


def temperature_name_from_index(  # pylint: disable=too-many-branches
    index: int,
    pvt_collector: pvt.PVT,
) -> TemperatureName:
    """
    Returns the temperature name from the index_handler.

    This method carries out a similar function in reverse to the `_get_index` method
    such that the temperature name, stored as a :class:`TemperatureName` instance, is
    returned based on the index passed in.

    :param index:
        The index of the temperature for which to return the temperature name.

    :param pvt_collector:
        The pvt collector being modelled.

    :return:
        The temperature name, as a :class:`TemperatureName` instance, based on the index
        passed in.

    :raises: LookupError
        Raised when an index is passed in for which no temperature name has been
        assigned in this function.

    """

    temperature_name: Optional[TemperatureName] = None

    # Determine the number of temperatures of each type.
    (
        num_upper_glass_temperatures,
        num_glass_temperatures,
        num_pv_temperatures,
        _,
        num_panel_temperatures,
        num_fluid_temperatures,
    ) = _calculate_number_of_temperatures(pvt_collector)

    if index < num_upper_glass_temperatures:
        temperature_name = TemperatureName.upper_glass
    elif index < num_upper_glass_temperatures + num_glass_temperatures:
        temperature_name = TemperatureName.glass
    elif (
        index
        < num_upper_glass_temperatures + num_glass_temperatures + num_pv_temperatures
    ):
        temperature_name = TemperatureName.pv
    elif index < num_panel_temperatures:
        temperature_name = TemperatureName.absorber
    elif index < num_panel_temperatures + num_fluid_temperatures:
        temperature_name = TemperatureName.pipe
    elif index < (num_panel_temperatures + 2 * num_fluid_temperatures):
        temperature_name = TemperatureName.htf
    elif index < (num_panel_temperatures + 3 * num_fluid_temperatures):
        temperature_name = TemperatureName.htf_in
    elif index < (num_panel_temperatures + 4 * num_fluid_temperatures):
        temperature_name = TemperatureName.htf_out
    elif index == (num_panel_temperatures + 4 * num_fluid_temperatures):
        temperature_name = TemperatureName.collector_in
    elif index == (num_panel_temperatures + 4 * num_fluid_temperatures + 1):
        temperature_name = TemperatureName.collector_out
    elif index == (num_panel_temperatures + 4 * num_fluid_temperatures + 2):
        temperature_name = TemperatureName.tank
    elif index == (num_panel_temperatures + 4 * num_fluid_temperatures + 3):
        temperature_name = TemperatureName.tank_in
    elif index == (num_panel_temperatures + 4 * num_fluid_temperatures + 4):
        temperature_name = TemperatureName.tank_out

    # Return the temperature name if assigned, else, return an error.
    if temperature_name is not None:
        return temperature_name
    raise LookupError(
        "A reverse lookup of the temperature name from the index was attempted: no "
        "matching temperature name was found."
    )


def x_coordinate(index: int, x_resolution: int) -> int:
    """
    Returns the x coordinate for the element being processed from the index_handler.

    :param index:
        The element index being processed.

    :param x_resolution:
        The x resolution of the simulation being run.

    :return:
        The x corodinate of the element.

    """

    return index % x_resolution


def y_coordinate(index: int, x_resolution: int) -> int:
    """
    Returns the y coordinate for the element being processed from the index_handler.

    :param index:
        The element index being processed.

    :param x_resolution:
        The x resolution of the simulation being run.

    :return:
        The y corodinate of the element.

    """

    return index // x_resolution



#!/usr/bin/python3.7
########################################################################################
# enforcement/mypy_enforcement.py
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2021
########################################################################################

"""
Script for ensuring that all mypy errors are declared correctly in the mypy yaml file.

All uses of the `type:ignore` mypy declaration must be declared in mypy_enforcement.yaml
in the enforcement folder.

"""

import os
import subprocess

from dataclasses import dataclass

import yaml

# String required in all enforcement entries.
JUSTIFICATION_STRING = "justification"
# Path to the YAML enforcement file.
YAML_ENFORCEMENT_FILE = os.path.join(
    "pvt_model", "enforcement", "mypy_enforcement.yaml"
)


class EnforcementError(Exception):
    """
    Raised when an enforcement error occurs.

    """

    def __init__(self, msg: str) -> None:
        """
        Instantiate a :class:`EnforcementError` instance.

        """

        super().__init__(
            f"Not all mypy 'type:ignore' uses were declared in the yaml file: {msg}"
        )


@dataclass
class TypeIgnoreUsage:
    """
    Represents a single use of the "type: ignore" declaration.

    .. attribute:: filename
        The name of the file containing the usage.

    .. attribute:: usage
        The actual usage within the file.

    """

    filename: str
    usage: str

    def __eq__(self, other) -> bool:
        """
        Determines if two instances are equal.

        """

        return str(self) == str(other)

    def __hash__(self) -> int:
        """
        Returns a unique string.

        """

        return hash(self.__repr__())

    def __lt__(self, other) -> bool:
        """
        Determines if the instance is less than the other.

        """

        return str(self) < str(other)

    def __repr__(self) -> str:
        """
        Representation of the type ignore usage.

        """

        return f"{self.filename}:{self.usage}"

    def __str__(self) -> str:
        """
        Representation of the type ignore usage.

        """

        return f"{self.filename}:{self.usage}"


def main() -> None:
    """
    Main method for the mypy enforcement module.

    """

    # Load the declared uses of "type:ignore".
    with open(YAML_ENFORCEMENT_FILE) as f:
        type_ignore_declarations = yaml.safe_load(f)

    try:
        processed_type_ignore_declarations = {
            TypeIgnoreUsage(entry["file"], entry["usage"])
            for entry in type_ignore_declarations
        }

    except KeyError as e:
        print(f"Not all data entries conform in the enforcement file: {str(e)}")
        raise

    if not all((JUSTIFICATION_STRING in entry for entry in type_ignore_declarations)):
        raise EnforcementError(
            "Not all entries were justified:\n  {}".format(
                "\n  ".join(
                    [
                        f"{entry['file']}: {entry['usage']}"
                        for entry in type_ignore_declarations
                        if JUSTIFICATION_STRING not in entry
                    ]
                )
            )
        )

    # Run from the above directory if needed.
    if "pvt_model" not in os.listdir():
        directory_prefix: str = ".." + os.path.sep
    else:
        directory_prefix = ""

    # Determine the actual uses
    type_ignore_uses = (
        subprocess.run(
            f"grep 'type: ignore' {directory_prefix}pvt_model -rn --exclude \\*.yaml",
            capture_output=True,
            check=True,
            shell=True,
        )
        .stdout.decode("utf-8")
        .split("\n")
    )
    type_ignore_uses = [
        entry.replace("# type: ignore", "") for entry in type_ignore_uses
    ]
    type_ignore_uses = [entry for entry in type_ignore_uses if entry != ""]

    # Process this into usable data.
    processed_type_ignore_uses = {
        TypeIgnoreUsage(
            entry.split(":")[0],
            ":".join(entry.split(":")[2:]).strip(),
        )
        for entry in type_ignore_uses
    }

    if processed_type_ignore_declarations != processed_type_ignore_uses:
        raise EnforcementError(
            "Missing declarations:\n{}\nExcess declarations:\n{}".format(
                "\n".join(
                    [
                        str(entry)
                        for entry in processed_type_ignore_uses
                        if entry not in processed_type_ignore_declarations
                    ]
                ),
                "\n".join(
                    [
                        str(entry)
                        for entry in processed_type_ignore_declarations
                        if entry not in processed_type_ignore_uses
                    ]
                ),
            )
        )


if __name__ == "__main__":
    main()



#!/usr/bin/python3.7
########################################################################################
# htf.py - The htf module for the matrix component.
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2021
########################################################################################

"""
The htf module for the matrix component.

This module computes and returns the equation(s) associated with the htf layer of the
PV-T collector for the matrix component.

The model works by arranging the system of differential equations as a matrix equation
such that the temperatures at each time step can be computed based on the coefficients
of the matrix which are computed based on the temperatures of the various components at
the previous time step, as well as various Physical and fundamental constants.

"""


from typing import List, Optional, Tuple, Union

import numpy

from .. import index_handler
from ..pvt_collector import pvt

from ...__utils__ import (
    OperatingMode,
    TemperatureName,
)
from ..physics_utils import density_of_water
from ..pvt_collector.element import Element

__all__ = ("calculate_htf_continuity_equation", "calculate_htf_equation")


def calculate_htf_continuity_equation(
    number_of_pipes: int,
    number_of_temperatures: int,
    number_of_x_elements: int,
    pvt_collector: pvt.PVT,
    element: Element,
) -> Tuple[List[float], float]:
    """
    Returns a matrix row and resultant vector value representing the htf equation.

    :param number_of_temperatures:
        The number of temperatures being modelled in the system.

    :return:
        A `tuple` containing:
        - the equation represented as a row in the matrix,
        - and the corresponding value in the resultant method.

    """

    # Compute the row equation
    row_equation: List[float] = [0] * number_of_temperatures

    # Compute the T_f(#, 0) term.
    row_equation[
        index_handler.index_from_pipe_coordinates(
            number_of_pipes,
            number_of_x_elements,
            element.pipe_index,  # type: ignore
            pvt_collector,
            TemperatureName.htf,
            element.y_index,
        )
    ] = 1

    # Compute the T_f,in(#, 0) term.
    row_equation[
        index_handler.index_from_pipe_coordinates(
            number_of_pipes,
            number_of_x_elements,
            element.pipe_index,  # type: ignore
            pvt_collector,
            TemperatureName.htf_in,
            element.y_index,
        )
    ] = -0.5

    # Compute the T_f,out(#, 0) term.
    row_equation[
        index_handler.index_from_pipe_coordinates(
            number_of_pipes,
            number_of_x_elements,
            element.pipe_index,  # type: ignore
            pvt_collector,
            TemperatureName.htf_out,
            element.y_index,
        )
    ] = -0.5

    return row_equation, 0


def calculate_htf_equation(
    best_guess_temperature_vector: Union[List[float], numpy.ndarray],
    number_of_pipes: int,
    number_of_temperatures: int,
    number_of_x_elements: int,
    operating_mode: OperatingMode,
    pipe_to_bulk_water_heat_transfer: float,
    previous_temperature_vector: Optional[numpy.ndarray],
    pvt_collector: pvt.PVT,
    resolution: Optional[int],
    element: Element,
) -> Tuple[List[float], float]:
    """
    Returns a matrix row and resultant vector value representing the htf equation.

    :param number_of_temperatures:
        The number of temperatures being modelled in the system.

    :return:
        A `tuple` containing:
        - the equation represented as a row in the matrix,
        - and the corresponding value in the resultant method.

    """

    # Compute the row equation
    row_equation: List[float] = [0] * number_of_temperatures

    if operating_mode.dynamic:
        bulk_water_internal_energy: float = (
            numpy.pi
            * (pvt_collector.absorber.inner_pipe_diameter / 2) ** 2  # [m^2]
            * element.length  # [m]
            * density_of_water(
                best_guess_temperature_vector[
                    index_handler.index_from_pipe_coordinates(
                        number_of_pipes,
                        number_of_x_elements,
                        element.pipe_index,  # type: ignore
                        pvt_collector,
                        TemperatureName.htf,
                        element.y_index,
                    )
                ]
            )  # [kg/m^3]
            * pvt_collector.absorber.htf_heat_capacity  # [J/kg*K]
            / resolution  # type: ignore  # [s]
        )  # [W/K]
    else:
        bulk_water_internal_energy = 0

    fluid_input_output_transfer_term = (
        pvt_collector.absorber.mass_flow_rate  # [kg/s]
        * pvt_collector.absorber.htf_heat_capacity  # [J/kg*K]
        / pvt_collector.absorber.number_of_pipes
    )  # [W/K]

    # Compute the T_f(#, j) term.
    row_equation[
        index_handler.index_from_pipe_coordinates(
            number_of_pipes,
            number_of_x_elements,
            element.pipe_index,  # type: ignore
            pvt_collector,
            TemperatureName.htf,
            element.y_index,
        )
    ] = (
        bulk_water_internal_energy + pipe_to_bulk_water_heat_transfer
    )

    # Compute the T_f,in(#, j) term.
    row_equation[
        index_handler.index_from_pipe_coordinates(
            number_of_pipes,
            number_of_x_elements,
            element.pipe_index,  # type: ignore
            pvt_collector,
            TemperatureName.htf_in,
            element.y_index,
        )
    ] = (
        -1 * fluid_input_output_transfer_term
    )

    # Compute the T_f,out(#, j) term.
    row_equation[
        index_handler.index_from_pipe_coordinates(
            number_of_pipes,
            number_of_x_elements,
            element.pipe_index,  # type: ignore
            pvt_collector,
            TemperatureName.htf_out,
            element.y_index,
        )
    ] = fluid_input_output_transfer_term

    # Compute the T_P(#, j) term.
    row_equation[
        index_handler.index_from_pipe_coordinates(
            number_of_pipes,
            number_of_x_elements,
            element.pipe_index,  # type: ignore
            pvt_collector,
            TemperatureName.pipe,
            element.y_index,
        )
    ] = (
        -1 * pipe_to_bulk_water_heat_transfer
    )

    # Compute the resultant vector value.
    resultant_vector_value = 0

    if operating_mode.dynamic:
        resultant_vector_value += (
            # Internal heat change.
            bulk_water_internal_energy
            * previous_temperature_vector[  # type: ignore
                index_handler.index_from_pipe_coordinates(
                    number_of_pipes,
                    number_of_x_elements,
                    element.pipe_index,  # type: ignore
                    pvt_collector,
                    TemperatureName.htf,
                    element.y_index,
                )
            ]
        )

    return row_equation, resultant_vector_value



#!/usr/bin/python3.7
########################################################################################
# argparser.py - The argument parser module for the PVT model.
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2021
########################################################################################

"""
The argument parser module for the PV-T model.

This module contains argument parsing code for the PVT module.

In a normal component, the argument parsing would happen within the main module.
However, the large number of arguments makes it neater for the code to be in a separate
module for the model.

"""

import argparse

from logging import Logger
from typing import Set

from .__utils__ import BColours, MissingParametersError, TemperatureName

__all__ = ("check_args", "parse_args")

# Used to keep track of the temperature layers based on the CLI arguments passsed in.
layer_map = {
    "dg": TemperatureName.upper_glass,
    "g": TemperatureName.glass,
    "pv": TemperatureName.pv,
    "a": TemperatureName.absorber,
    "p": TemperatureName.pipe,
    "f": TemperatureName.htf,
}


class ArgumentMismatchError(Exception):
    """
    Raised when arguments passed in mismatch.

    """

    def __init__(self, msg: str) -> None:
        """
        Instantiate a :class:`ArgumentMismatchError`.

        :param msg:
            The message to append to the output.

        """

        super().__init__(f"Mismatch in command-line arguments: {msg}")


def check_args(  # pylint: disable=too-many-branches
    parsed_args: argparse.Namespace, logger: Logger, number_of_pipes: int
) -> Set[TemperatureName]:
    """
    Enforces rules on the command-line arguments passed in in addition to argparse rules

    :param parsed_args:
        The parsed command-line arguments.

    :param logger:
        The logger used for the run.

    :param number_of_pipes:
        The number of pipes attached to the absorber being modelled.

    :raises: ArgumentMismatchError
        Raised if the command-line arguments mismatch.

    """

    # Check that the output file is correctly specified without a file extension.
    if parsed_args.output is None or parsed_args.output == "":
        logger.error(
            "%sAn output filename must be provided on the command-line interface.%s",
            BColours.FAIL,
            BColours.ENDC,
        )
        raise MissingParametersError(
            "Command-Line Interface", "An output file name must be provided."
        )
    if parsed_args.output.endswith(".yaml") or parsed_args.output.endswith(".json"):
        logger.error(
            "%sThe output filename must be irrespective of data type..%s",
            BColours.FAIL,
            BColours.ENDC,
        )
        raise Exception(
            "The output file must be irrespecitve of file extension/data type."
        )

    # Enforce that all required arguments are specified.
    if parsed_args.portion_covered is None:
        raise ArgumentMismatchError(
            "{}The argument `--portion-covered` must be used when running ".format(
                BColours.FAIL
            )
            + "the model.{}".format(BColours.ENDC)
        )

    if parsed_args.initial_month is None:
        raise ArgumentMismatchError(
            "{}The argument `--initial-month` must be used when running ".format(
                BColours.FAIL
            )
            + "the model to specify the weather data to use.{}".format(BColours.ENDC)
        )

    if parsed_args.output is None:
        raise ArgumentMismatchError(
            "{}The argument `--output` must be used when running the model to".format(
                BColours.FAIL
            )
            + "specify the name of the output file.{}".format(BColours.ENDC)
        )

    if parsed_args.pvt_data_file is None:
        raise ArgumentMismatchError(
            "{}The argument `--pvt-data-file` must be used when running ".format(
                BColours.FAIL
            )
            + "the model to specify the PVT YAML file.{}".format(BColours.ENDC)
        )

    # Enforce the resolution requirements.
    if parsed_args.x_resolution == 1 and parsed_args.y_resolution == 1:
        logger.warn(
            "%s1x1 resolution is depreciated, consider running at a higher resolution."
            "%s",
            BColours.FAIL,
            BColours.ENDC,
        )
    elif (
        parsed_args.x_resolution >= (2 * number_of_pipes + 3)
        # and parsed_args.x_resolution % 2 == 1
        and parsed_args.y_resolution >= 3
    ):
        logger.info(
            "Resolution of %s by %s is accpetable and is greater than %s by 3.",
            parsed_args.x_resolution,
            parsed_args.y_resolution,
            int((2 * number_of_pipes + 3)),
        )
    else:
        logger.error(
            "%sThe specified resolution of %s by %s is not supported. The resolution "
            "must be either 1 by 1 or greater than %s by 3.%s",
            BColours.FAIL,
            parsed_args.x_resolution,
            parsed_args.y_resolution,
            int((2 * number_of_pipes + 3)),
            BColours.ENDC,
        )
        raise ArgumentMismatchError(
            "The specified resolution of {} by {} is not supported. The ".format(
                parsed_args.x_resolution, parsed_args.y_resolution
            )
            + "resolution must be either 1 by 1 or greater than {} by 3".format(
                int((2 * number_of_pipes + 3))
            )
        )

    # Enforce the matching up of dynamic and steady-state arguments.
    # Enforce that either dynamic or steady-state is specified, but not both.
    if parsed_args.dynamic is parsed_args.steady_state:
        raise ArgumentMismatchError(
            "{}Either `--dynamnic` or `--steady-state` should be used, but ".format(
                BColours.FAIL
            )
            + "not both.{}".format(BColours.ENDC)
        )

    # Enforce that, if decoupled is specified, steady-state is specified.
    if not parsed_args.decoupled and (
        not parsed_args.dynamic or parsed_args.steady_state
    ):
        raise ArgumentMismatchError(
            "{}If `--decoupled` is not specified, the system must be run in ".format(
                BColours.FAIL
            )
            + "dynamic mode.{}".format(BColours.ENDC)
        )

    # Enforce that, if decoupled is specified, solar irradiance is specified.
    if parsed_args.steady_state and not parsed_args.steady_state_data_file:
        raise ArgumentMismatchError(
            "{}If `--steady_state` is specified, the steady-state data file must ".format(
                BColours.FAIL
            )
            + "be specified with `--steady-state-data-file`.{}".format(BColours.ENDC)
        )

    # Enforce that, if decoupled is not specified, either days or months is specified.
    if not parsed_args.steady_state and not (
        parsed_args.months is not None
        or parsed_args.days is not None
        or parsed_args.minutes is not None
    ):
        raise ArgumentMismatchError(
            "{}If running a coupled system, the number of days or months for ".format(
                BColours.FAIL
            )
            + "the run must be specified with `--days` or `--months`.{}".format(
                BColours.ENDC
            )
        )

    # Enforce that, if minutes is specified, that it is greater than the resolution.
    if parsed_args.minutes is not None:
        if parsed_args.resolution > parsed_args.minutes * 60:
            raise ArgumentMismatchError(
                "{}The resolution of the simulation must be less than the time ".format(
                    BColours.FAIL
                )
                + "for which it is being run.{}".format(BColours.ENDC)
            )

    # Enforce that, if decoupled is not specified, start-time is specified.
    if not parsed_args.decoupled and parsed_args.start_time is None:
        raise ArgumentMismatchError(
            "{}If running a coupled system, the start time for the run must be ".format(
                BColours.FAIL
            )
            + "specified with `--start-time`.{}".format(BColours.ENDC)
        )

    # Enforce the layer names are of the correct type.
    if not all((entry in layer_map for entry in parsed_args.layers)):
        raise ArgumentMismatchError(
            "{}If using the --layers developer argument, only layers {} can be "
            "specified. Rogue layer name: '{}'.{}".format(
                BColours.FAIL,
                ", ".join(layer_map.keys()),
                ", ".join(
                    [entry for entry in parsed_args.layers if entry not in layer_map]
                ),
                BColours.ENDC,
            )
        )

    # Enforce that, if double glazing is specified, then the glass layer is also present
    if "dg" in parsed_args.layers and "g" not in parsed_args.layers:
        raise ArgumentMismatchError(
            "{}If using the --layers developer argument, the glass layer, ".format(
                BColours.FAIL
            )
            + "'g', must be specified if using double-glazing, 'dg'.{}".format(
                BColours.ENDC,
            )
        )

    return {layer_map[entry] for entry in parsed_args.layers}


def parse_args(args) -> argparse.Namespace:
    """
    Parse command-line arguments.

    :param args:
        The command-line arguments to parse.

    :return:
        The parsed arguments in an :class:`argparse.Namespace`.

    """

    parser = argparse.ArgumentParser()
    developer_arguments = parser.add_argument_group("developer arguments")
    dynamic_arguments = parser.add_argument_group("dynamic arguments")
    required_named_arguments = parser.add_argument_group("required named arguments")
    steady_state_arguments = parser.add_argument_group("steady-state arguments")

    steady_state_arguments.add_argument(
        "--ambient-temperature",
        "-at",
        type=float,
        help="[decoupled] The ambient temperature surrounding the absorber, in "
        "degrees Celcius, needs to be specified if running a decoupled system.",
    )
    parser.add_argument(
        "--average-irradiance",
        "-ar",
        action="store_true",
        default=False,
        help="Whether to average the solar irradiance intensities on a monthly basis, "
        "or to use the data for each day individually.",
    )
    steady_state_arguments.add_argument(
        "--collector-input-temperature",
        "-ci",
        default=None,
        type=float,
        help="[decoupled] The input temperature in degrees Celcius of HTF to use when "
        "modelling a decoupled PVT absorber.",
    )
    parser.add_argument(
        "--cloud-efficacy-factor",
        "-c",
        type=float,
        help="The effect that the cloud cover has, rated between 0 (no effect) and 1.",
    )
    dynamic_arguments.add_argument(
        "--days",
        "-d",
        type=int,
        help="The number of days to run the simulation for. Overrides 'months'.",
    )
    parser.add_argument(
        "--decoupled",
        action="store_true",
        default=False,
        help="If specified, the model will be run with a decoupled PVT absorber.",
    )
    parser.add_argument(
        "--disable-logging",
        action="store_true",
        default=False,
        help="If specified, logging will be disabled.",
    )
    parser.add_argument(
        "--dynamic",
        default=False,
        action="store_true",
        help="If specified, Fourier number calculation will be skipped and a dynamic "
        "model will be used for the run.",
    )
    dynamic_arguments.add_argument(
        "--exchanger-data-file",
        "-e",
        help="The location of the Exchanger system YAML data file.",
    )
    dynamic_arguments.add_argument(
        "--initial-month",
        "-i",
        type=int,
        help="The first month for which the simulation will be run, expressed as an "
        "int. The default is 1, corresponding to January.",
    )
    parser.add_argument(
        "--initial-system-temperature-vector",
        nargs="+",
        help="If specified, this will override the internal initial system "
        "temperature vector.",
    )
    developer_arguments.add_argument(
        "--layers",
        default=layer_map.keys(),
        nargs="+",
        type=str,
        help="Used to specifiy layers present. Options are 'g', 'pv', 'a', 'p', and "
        "'f'.",
    )
    required_named_arguments.add_argument(
        "--location", "-l", help="The location for which to run the simulation."
    )
    parser.add_argument(
        "--mass-flow-rate",
        type=float,
        default=None,
        help="Can be used to override the mass-flow rate used in the collector.",
    )
    dynamic_arguments.add_argument(
        "--minutes",
        "-min",
        help="Can be used to specify only a small number of minutes for which to run the model.",
        type=int,
    )
    dynamic_arguments.add_argument(
        "--months",
        "-m",
        help="The number of months for which to run the simulation. Default is 12.",
        default=12,
        type=int,
    )
    parser.add_argument(
        "--number-of-people",
        "-n",
        type=int,
        help="The number of household members to consider in this model.",
    )
    required_named_arguments.add_argument(
        "--output",
        "-o",
        help="The output file to save data to. This should be of JSON format.",
    )
    required_named_arguments.add_argument(
        "--portion-covered",
        "-pc",
        type=float,
        help="The portion of the panel which is covered in PV, "
        "from 1 (all) to 0 (none).",
    )
    dynamic_arguments.add_argument(
        "--pump-data-file",
        "-pm",
        help="The location of the pump YAML data file for the PV-T system pump.",
    )
    required_named_arguments.add_argument(
        "--pvt-data-file", "-p", help="The location of the PV-T system YAML data file."
    )
    dynamic_arguments.add_argument(
        "--resolution",
        "-r",
        help="The resolution, in seconds, used to solve the panel temperatures. "
        "Defaults to one minute.",
        type=int,
        default=60,
    )
    parser.add_argument(
        "--return-system-data",
        help=argparse.SUPPRESS,
        default=False,
        action="store_true",
    )
    parser.add_argument(
        "--skip-2d-output",
        "-sk2d",
        action="store_true",
        default=False,
        help="If specified, the 2D output will not be saved, and only 1D info, and "
        "plots, will be saved and generated.",
    )
    parser.add_argument(
        "--skip-analysis",
        "-sa",
        action="store_true",
        default=False,
    )
    steady_state_arguments.add_argument(
        "--solar-irradiance",
        "-sol",
        default=None,
        type=float,
        help="[decoupled] The solar irradiance in Watts per meter squared to use when "
        "running the system as a decoupled PVT absorber.",
    )
    dynamic_arguments.add_argument(
        "--start-time",
        "-st",
        type=int,
        default=0,
        help="The start time, in hours, at which to begin the simulation during the day",
    )
    parser.add_argument(
        "--steady-state",
        "-ss",
        action="store_true",
        default=False,
        help="If specified, Fourier-number calculation will be skipped and the model "
        "will be run without transient/dynamic terms.",
    )
    steady_state_arguments.add_argument(
        "--steady-state-data-file",
        "-ssdf",
        default=None,
        type=str,
        help="The path to the data file containing information specifying the steady-"
        "state runs that should be carried out.",
    )
    dynamic_arguments.add_argument(
        "--tank-data-file",
        "-t",
        help="The location of the Hot-Water Tank system YAML data file.",
    )
    parser.add_argument(
        "--use-pvgis",
        default=False,
        action="store_true",
        help="If specified, PVGIS data is used. Otherwise, the data extracted from "
        "Maria's paper is used.",
    )
    parser.add_argument(
        "--unglazed",
        "-u",
        help="If specified, the panel will be un-glazed, i.e., without a glass coating.",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--verbose",
        "-v",
        action="store_true",
        default=False,
        help="If specified, verbose logging will be carried out.",
    )
    parser.add_argument(
        "--wind-speed",
        "-w",
        default=None,
        help="If specified, overrides the wind-speed profiles used.",
        type=int,
    )
    required_named_arguments.add_argument(
        "--x-resolution",
        "-x",
        help="The number of elements to include in the x direction (across the panel) "
        "for the simulation.",
        type=int,
    )
    required_named_arguments.add_argument(
        "--y-resolution",
        "-y",
        help="The number of elements to include in the y direction (along the length of "
        "the panel) for the simulation.",
        type=int,
    )

    return parser.parse_args(args)



#!/usr/bin/python3.7
########################################################################################
# pvt_collector/__utils__.py - The utility module for the PVT panel component.
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2021
########################################################################################

"""
The utility module for the PV-T panel component.

This module contains common functionality, strucutres, and types, to be used by the
various modules throughout the PVT panel component.

"""

from dataclasses import dataclass
from math import sqrt

from ..__utils__ import (
    LayerParameters,
    OpticalLayerParameters,
)

__all__ = (
    "Layer",
    "OpticalLayer",
)


class Layer:
    """
    Represents a layer within the PV-T panel.

    .. attribute:: conductivity
        The conductivity of the layer, measured in Watts per meter Kelvin.

    .. attribute:: density
        The desntiy of the layer, measured in kilograms per meter cubed.

    .. attribute:: heat_capacity
      The heat capacity of the layer, measured in Joules per kilogram Kelvin.

    .. attribute:: thickenss
        The thickness (depth) of the layer, measured in meters.

    """

    def __init__(self, layer_params: LayerParameters) -> None:
        """
        Instantiate an instance of the layer class.

        """

        self.conductivity = layer_params.conductivity
        self.density = layer_params.density
        self.heat_capacity = layer_params.heat_capacity  # [J/kg*K]
        self.thickness = layer_params.thickness


@dataclass
class MicroLayer:
    """
    Represents a layer within the panel which is small enough to ignore dynamic terms.

    Such layers, such as the adhesive between two layers, are too small to be treated in
    the same way as actual layers as their dynamic (heat capacity) terms are too small
    to warrant consideration.

    .. attribute:: conductivity
        The conductivity of the layer, measured in Watts per meter Kelvin.

    .. attribute:: thickenss
        The thickness (depth) of the layer, measured in meters.

    """

    conductivity: float
    thickness: float


class OpticalLayer(Layer):
    """
    Represents a layer within the PV-T panel that has optical properties.

    .. attribute:: absorptivity
        The absorptivity of the layer: a dimensionless number between 0 (nothing is
        absorbed by the layer) and 1 (all light is absorbed).

    .. attribute:: emissivity
        The emissivity of the layer; a dimensionless number between 0 (nothing is
        emitted by the layer) and 1 (the layer re-emits all incident light).

    .. attribute:: transmissivity
        The transmissivity of the layer: a dimensionless number between 0 (nothing is
        transmitted through the layer) and 1 (all light is transmitted).

    """

    def __init__(self, optical_params: OpticalLayerParameters) -> None:
        """
        Instantiate an optical layer within the PV-T panel.

        :param optical_params:
            Contains parameters needed to instantiate the optical layer.

        """

        super().__init__(
            LayerParameters(
                optical_params.conductivity,
                optical_params.density,
                optical_params.heat_capacity,
                optical_params.thickness,
            )
        )

        self.absorptivity = optical_params.absorptivity
        self.transmissivity = optical_params.transmissivity
        self.emissivity = optical_params.emissivity

    def __repr__(self) -> str:
        """
        Returns a nice representation of the layer.

        :return:
            A `str` giving a nice representation of the layer.

        """

        return (
            "OpticalLayer("
            f"absorptivity: {self.absorptivity}, "
            f"emissivitiy: {self.emissivity}, "
            f"heat_capacity: {self.heat_capacity}, "
            f"reflectivity: {self.reflectivity:.3f}, "
            f"thickness: {self.thickness}, "
            f"transmissivity: {self.transmissivity}"
            ")"
        )

    @property
    def reflectance(self) -> float:
        """
        Returns the reflectance of the layer.

        :return:
            The reflectance of the layer.

        """

        return self.reflectivity ** 2

    @property
    def reflectivity(self) -> float:
        """
        Returns the reflectivity of the layer.

        :return:
            The reflectivity of the layer, based on its other optical properties.

        """

        return sqrt(1 - self.absorptivity ** 2 - self.transmissivity ** 2)

    @property
    def transmittance(self) -> float:
        """
        Returns the transmittance of the layer.

        :return:
            The transmittance of the layer.

        """

        return self.transmissivity ** 2



#!/usr/bin/python3.7
########################################################################################
# absorber.py - The absorber module for the matrix component.
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2021
########################################################################################

"""
The absorber module for the matrix component.

This module computes and returns the equation(s) associated with the absorber layer of
the PV-T collector for the matrix component.

The model works by arranging the system of differential equations as a matrix equation
such that the temperatures at each time step can be computed based on the coefficients
of the matrix which are computed based on the temperatures of the various components at
the previous time step, as well as various Physical and fundamental constants.

"""

import logging

from typing import List, Optional, Tuple, Union

import numpy

from .. import index_handler, physics_utils
from ..pvt_collector import pvt

from ...__utils__ import OperatingMode, ProgrammerJudgementFault, TemperatureName
from ..__utils__ import WeatherConditions
from ..pvt_collector.element import Element, ElementCoordinates
from ..pvt_collector.physics_utils import insulation_thermal_resistance


__all__ = ("calculate_absorber_equation",)


def calculate_absorber_equation(  # pylint: disable=too-many-branches
    absorber_to_pipe_conduction: float,
    best_guess_temperature_vector: Union[List[float], numpy.ndarray],
    glass_downward_conduction: float,
    glass_downward_radiation: float,
    logger: logging.Logger,
    number_of_pipes: int,
    number_of_temperatures: int,
    number_of_x_elements: int,
    number_of_y_elements: int,
    operating_mode: OperatingMode,
    previous_temperature_vector: Optional[numpy.ndarray],
    pv_to_absorber_conduction: float,
    pvt_collector: pvt.PVT,
    resolution: Optional[int],
    element: Element,
    weather_conditions: WeatherConditions,
) -> Tuple[List[float], float]:
    """
    Returns a matrix row and resultant vector value representing the absorber equation.

    :param number_of_temperatures:
        The number of temperatures being modelled in the system.

    :return:
        A `tuple` containing:
        - the equation represented as a row in the matrix,
        - and the corresponding value in the resultant method.

    """

    logger.debug(
        "Beginning calculation of absorber equation for element %s.",
        element.coordinates,
    )

    # Compute the row equation
    row_equation: List[float] = [0] * number_of_temperatures

    # import pdb

    # pdb.set_trace(header=f"T_A{element.coordinates}")

    if operating_mode.dynamic:
        collector_internal_energy_change: float = (
            element.width  # [m]
            * element.length  # [m]
            * pvt_collector.absorber.thickness  # [m]
            * pvt_collector.absorber.density  # [kg/m^3]
            * pvt_collector.absorber.heat_capacity  # [J/kg*K]
            / resolution  # type: ignore  # [s]
        )
    else:
        collector_internal_energy_change = 0
    logger.debug(
        "Absorber internal energy term: %s W/K", collector_internal_energy_change
    )

    # Compute the positive conductive term based on the next element along.
    positive_x_element = pvt_collector.elements.get(
        ElementCoordinates(element.x_index + 1, element.y_index)
    )
    if positive_x_element is not None:
        positive_x_wise_conduction: float = (
            pvt_collector.absorber.conductivity  # [W/m*K]
            * pvt_collector.absorber.thickness  # [m]
            * element.length  # [m]
            / (0.5 * (element.width + positive_x_element.width))  # [m]
        )
    else:
        positive_x_wise_conduction = 0
    logger.debug(
        "Positive absorber x-wise conduction term: %s W/K", positive_x_wise_conduction
    )

    # Compute the positive conductive term based on the next element along.
    negative_x_element = pvt_collector.elements.get(
        ElementCoordinates(element.x_index - 1, element.y_index)
    )
    if negative_x_element is not None:
        negative_x_wise_conduction: float = (
            pvt_collector.absorber.conductivity  # [W/m*K]
            * pvt_collector.absorber.thickness  # [m]
            * element.length  # [m]
            / (0.5 * (element.width + negative_x_element.width))  # [m]
        )
    else:
        negative_x_wise_conduction = 0
    logger.debug(
        "Negative absorber x-wise conduction term: %s W/K", negative_x_wise_conduction
    )

    # Compute the overall x-wise conduction term.
    x_wise_conduction = positive_x_wise_conduction + negative_x_wise_conduction
    logger.debug("Absorber x-wise conduction term: %s W/K", x_wise_conduction)

    # Compute the positive conductive term based on the next element along.
    positive_y_element = pvt_collector.elements.get(
        ElementCoordinates(element.x_index, element.y_index + 1)
    )
    if positive_y_element is not None:
        positive_y_wise_conduction: float = (
            pvt_collector.absorber.conductivity  # [W/m*K]
            * pvt_collector.absorber.thickness  # [m]
            * element.width  # [m]
            / (0.5 * (element.length + positive_y_element.length))  # [m]
        )
    else:
        positive_y_wise_conduction = 0
    logger.debug(
        "Positive absorber y-wise conduction term: %s W/K", positive_y_wise_conduction
    )

    # Compute the positive conductive term based on the next element along.
    negative_y_element = pvt_collector.elements.get(
        ElementCoordinates(element.x_index, element.y_index - 1)
    )
    if negative_y_element is not None:
        negative_y_wise_conduction: float = (
            pvt_collector.absorber.conductivity  # [W/m*K]
            * pvt_collector.absorber.thickness  # [m]
            * element.width  # [m]
            / (0.5 * (element.length + negative_y_element.length))  # [m]
        )
    else:
        negative_y_wise_conduction = 0
    logger.debug(
        "Negative absorber y-wise conduction term: %s W/K", negative_y_wise_conduction
    )

    # Compute the overall y-wise conduction term.
    y_wise_conduction = positive_y_wise_conduction + negative_y_wise_conduction
    logger.debug("Absorber y-wise conduction term: %s W/K", y_wise_conduction)

    absorber_to_insulation_loss = (
        (
            element.width  # [m]
            * element.length  # [m]
            / insulation_thermal_resistance(
                best_guess_temperature_vector,
                pvt_collector,
                index_handler.index_from_element_coordinates(
                    number_of_x_elements,
                    pvt_collector,
                    TemperatureName.absorber,
                    element.x_index,
                    element.y_index,
                ),
                weather_conditions,
            )  # [m^2*K/W]
        )
        if not element.pipe
        else 0
    )
    logger.debug(
        "Absorber to insulation loss term: %s W/K", absorber_to_insulation_loss
    )

    # If there are no upper layers, compute the upward loss terms.
    if not element.pv and not element.glass:
        (
            absorber_to_air_conduction,
            absorber_to_sky_radiation,
        ) = physics_utils.upward_loss_terms(
            best_guess_temperature_vector,
            pvt_collector,
            element,
            pvt_collector.absorber.emissivity,
            index_handler.index_from_element_coordinates(
                number_of_x_elements,
                pvt_collector,
                TemperatureName.absorber,
                element.x_index,
                element.y_index,
            ),
            weather_conditions,
        )
        logger.debug("Absorber to air conduction %s W/K", absorber_to_air_conduction)
        logger.debug("Absorber to sky radiation %s W/K", absorber_to_sky_radiation)
    else:
        absorber_to_air_conduction = 0
        absorber_to_sky_radiation = 0

    # Compute the T_A(i, j) term
    row_equation[
        index_handler.index_from_element_coordinates(
            number_of_x_elements,
            pvt_collector,
            TemperatureName.absorber,
            element.x_index,
            element.y_index,
        )
    ] = (
        collector_internal_energy_change
        + x_wise_conduction
        + y_wise_conduction
        + (pv_to_absorber_conduction if element.pv else 0)
        + absorber_to_pipe_conduction
        + absorber_to_insulation_loss
        + (
            absorber_to_air_conduction + absorber_to_sky_radiation
            if (not element.pv and not element.glass)
            else 0
        )
        + (
            glass_downward_conduction + glass_downward_radiation
            if (element.glass and not element.pv)
            else 0
        )
    )

    # Compute the T_A(i+1, j) term provided that that element exists.
    if element.x_index + 1 < number_of_x_elements:
        row_equation[
            index_handler.index_from_element_coordinates(
                number_of_x_elements,
                pvt_collector,
                TemperatureName.absorber,
                element.x_index + 1,
                element.y_index,
            )
        ] = (
            -1 * positive_x_wise_conduction
        )

    # Compute the T_A(i-1, j) term provided that that element exists.
    if element.x_index > 0:
        row_equation[
            index_handler.index_from_element_coordinates(
                number_of_x_elements,
                pvt_collector,
                TemperatureName.absorber,
                element.x_index - 1,
                element.y_index,
            )
        ] = (
            -1 * negative_x_wise_conduction
        )

    # Compute the T_A(i, j+1) term provided that that element exists.
    if element.y_index + 1 < number_of_y_elements:
        row_equation[
            index_handler.index_from_element_coordinates(
                number_of_x_elements,
                pvt_collector,
                TemperatureName.absorber,
                element.x_index,
                element.y_index + 1,
            )
        ] = (
            -1 * positive_y_wise_conduction
        )

    # Compute the T_A(i, j-1) term provided that that element exists.
    if element.y_index > 0:
        row_equation[
            index_handler.index_from_element_coordinates(
                number_of_x_elements,
                pvt_collector,
                TemperatureName.absorber,
                element.x_index,
                element.y_index - 1,
            )
        ] = (
            -1 * negative_y_wise_conduction
        )

    # Compute the T_pv(i, j) term provided that there is a absorber layer present.
    if element.pv:
        row_equation[
            index_handler.index_from_element_coordinates(
                number_of_x_elements,
                pvt_collector,
                TemperatureName.pv,
                element.x_index,
                element.y_index,
            )
        ] = (
            -1 * pv_to_absorber_conduction
        )
    elif element.glass:
        row_equation[
            index_handler.index_from_element_coordinates(
                number_of_x_elements,
                pvt_collector,
                TemperatureName.glass,
                element.x_index,
                element.y_index,
            )
        ] = -1 * (glass_downward_conduction + glass_downward_radiation)

    # Compute the T_P(pipe_number, j) term provided that there is a pipe present.
    if element.pipe:
        if element.pipe_index is None:
            raise ProgrammerJudgementFault(
                "A element specifies a pipe is present yet no pipe index is supplied. "
                f"Element: {str(element)}"
            )
        try:
            row_equation[
                index_handler.index_from_pipe_coordinates(
                    number_of_pipes,
                    number_of_x_elements,
                    element.pipe_index,
                    pvt_collector,
                    TemperatureName.pipe,
                    element.y_index,
                )
            ] = (
                -1 * absorber_to_pipe_conduction
            )
        except ProgrammerJudgementFault as e:
            raise ProgrammerJudgementFault(
                "Error determining pipe temperature term. "
                f"Likely that pipe index was not specified: {str(e)}"
            ) from None

    # Compute the solar-thermal absorption if relevant.
    if not element.pv:
        solar_thermal_resultant_vector_absorbtion_term = (
            pvt_collector.absorber_transmissivity_absorptivity_product
            * weather_conditions.irradiance  # [W/m^2]
            * element.width  # [m]
            * element.length  # [m]
        )
    else:
        solar_thermal_resultant_vector_absorbtion_term = 0

    # Compute the resultant vector value.
    resultant_vector_value = (
        # Ambient temperature term.
        absorber_to_insulation_loss  # [W/K]
        * weather_conditions.ambient_temperature  # [K]
        # Solar absorption term.
        + solar_thermal_resultant_vector_absorbtion_term  # [W]
        # Heat loss to the air.
        + absorber_to_air_conduction * weather_conditions.ambient_temperature
        # Absorber to sky radiation
        + absorber_to_sky_radiation * weather_conditions.sky_temperature
    )

    if operating_mode.dynamic:
        resultant_vector_value += (
            # Internal heat change term.
            collector_internal_energy_change  # [W/K]
            * previous_temperature_vector[  # type: ignore
                index_handler.index_from_element_coordinates(
                    number_of_x_elements,
                    pvt_collector,
                    TemperatureName.absorber,
                    element.x_index,
                    element.y_index,
                )
            ]  # [K]
        )

        try:
            logger.debug(
                "Rough Absorber Temperature estimate: %s K.",
                int(
                    resultant_vector_value
                    / (absorber_to_insulation_loss + collector_internal_energy_change)
                ),
            )
        except ZeroDivisionError:
            logger.debug(
                "Absorber temperature estimate could not be computed due to zero-division."
            )

    return row_equation, resultant_vector_value



#!/usr/bin/python3.7
########################################################################################
# pvt_collector/pv.py - Represents a absorber within a PVT panel.
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2021
########################################################################################

"""
The PV module for the PV-T model.

This module represents a PV layer within a PV-T panel.

"""

from ..__utils__ import (
    OpticalLayerParameters,
    PVParameters,
)
from .__utils__ import (
    OpticalLayer,
)

__all__ = ("PV",)


class PV(OpticalLayer):
    """
    Represents the photovoltaic (middle) layer of the PV-T panel.

    """

    # Private attributes:
    #
    # .. attribute:: reference_efficiency
    #   The efficiency of the PV layer at the reference temperature. Thie value varies
    #   between 1 (corresponding to 100% efficiency), and 0 (corresponding to 0%
    #   efficiency)
    #
    # .. attribute:: reference_temperature
    #   The referencee temperature, in Kelvin, at which the reference efficiency is
    #   defined.
    #
    # .. attribute:: thermal_coefficient
    #   The thermal coefficient for the efficiency of the panel.
    #

    def __init__(self, pv_params: PVParameters) -> None:
        """
        Instantiate a PV layer.

        :param pv_params:
            Parameters needed to instantiate the PV layer.

        """

        super().__init__(
            OpticalLayerParameters(
                pv_params.conductivity,
                pv_params.density,
                pv_params.heat_capacity,
                pv_params.thickness,
                pv_params.transmissivity,
                pv_params.absorptivity,
                pv_params.emissivity,
            )
        )

        self.reference_efficiency = pv_params.reference_efficiency
        self.reference_temperature = pv_params.reference_temperature
        self.thermal_coefficient = pv_params.thermal_coefficient

    def __repr__(self) -> str:
        """
        Returns a nice representation of the layer.

        :return:
            A `str` giving a nice representation of the layer.

        """

        return (
            "PV("
            f"absorptivity: {self.absorptivity}, "
            f"heat_capacity: {self.heat_capacity}J/kg*K, "
            f"transmissivity: {self.transmissivity}, "
            f"emissivity: {self.emissivity}, "
            f"reference_efficiency: {self.reference_efficiency}, "
            f"reference_temperature: {self.reference_temperature}K, "
            f"reflectivity: {self.reflectivity:.3f}, "
            f"thermal_coefficient: {self.thermal_coefficient}K^(-1), "
            f"thickness: {self.thickness}m"
            ")"
        )

    def electrical_efficiency(self, pv_temperature: float) -> float:
        """
        Returns the electrical efficiency of the PV panel based on its temperature.

        :param pv_temperature:
            The temperature of the PV layer, measured in Kelvin.

        :return:
            A decimal giving the percentage efficiency of the PV panel between 0 (0%
            efficiency), and 1 (100% efficiency).

        """

        return self.reference_efficiency * (  # [unitless]
            1
            - self.thermal_coefficient  # [1/K]
            * (pv_temperature - self.reference_temperature)  # [K]
        )



#!/usr/bin/python3.7
########################################################################################
# physics_utils.py - Utility module containing Physics equations.
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2021
########################################################################################

"""
The physics utility module for the PVT model.

"""

from typing import List, Optional, Tuple, Union

import math

from numpy import ndarray, pi

from .constants import (
    ACCELERATION_DUE_TO_GRAVITY,
    STEFAN_BOLTZMAN_CONSTANT,
    THERMAL_CONDUCTIVITY_OF_WATER,
    ZERO_CELCIUS_OFFSET,
)
from .pvt_collector import pvt
from .pvt_collector.element import Element

from .__utils__ import ProgrammerJudgementFault, WeatherConditions

__all__ = (
    "convective_heat_transfer_coefficient_of_water",
    "density_of_water",
    "dynamic_viscosity_of_water",
    "free_heat_transfer_coefficient_of_air",
    "grashof_number",
    "prandtl_number",
    "radiative_heat_transfer_coefficient",
    "rayleigh_number",
    "reduced_temperature",
    "reynolds_number",
    "upward_loss_terms",
)


def _top_heat_transfer_coefficient(
    pvt_collector: pvt.PVT,
    element_top_temperature: float,
    weather_conditions: WeatherConditions,
) -> float:
    """
    Computes the heat-transfer coeffient between the top of the panel and the air.

    NOTE: This includes both conductive (free) and convective (forced) heat transfers.

    :param pvt_collector:
        The pvt panel being modelled.

    :param element_top_temperature:
        The temperature, measured in Kelvin, of the top-layer of the element.

    :param weather_conditions:
        The weather conditions at the time step being modelled.

    :return:
        The heat transfer coefficient between the panel and the air, measured in Watts
        per meter squared Kelvin.

    """

    heat_transfer_coefficient: float = (
        weather_conditions.wind_heat_transfer_coefficient ** 3
        + free_heat_transfer_coefficient_of_air(
            pvt_collector, element_top_temperature, weather_conditions
        )
        ** 3
    ) ** (1 / 3)

    return heat_transfer_coefficient


def convective_heat_transfer_coefficient_of_water(
    fluid_temperature: float,
    pvt_collector: pvt.PVT,
    weather_conditions: WeatherConditions,
) -> float:
    """
    Computes the convective heat transfer to a fluid in Watts.

    :param fluid_temperature:
        The temperature of the fluid element, measured in Kelvin.

    :param pvt_collector:
        The PVT panel being modelled.

    :param weather_conditions:
        The weather conditions at the current time step.

    :return:
        The convective heat transfer to the fluid, measured in Watts. If the value is
        positive, then the heat flow is from the container walls to the fluid. If the
        value returned is negative, then the flow is from the fluid to the container
        walls.

    """

    # Convert the mass-flow rate into a flow speed.
    flow_speed: float = pvt_collector.absorber.mass_flow_rate / (
        density_of_water(fluid_temperature)
        * pi
        * (pvt_collector.absorber.inner_pipe_diameter / 2) ** 2
    )

    # Compute the current Reynolds number.
    current_reynolds_number = reynolds_number(
        density_of_water(fluid_temperature),
        dynamic_viscosity_of_water(fluid_temperature),
        flow_speed,
        pvt_collector.absorber.inner_pipe_diameter,
    )

    if flow_speed == 0:
        return (
            2
            * THERMAL_CONDUCTIVITY_OF_WATER
            / pvt_collector.absorber.inner_pipe_diameter
        )

    if current_reynolds_number < 2300:
        return (
            4.36
            * THERMAL_CONDUCTIVITY_OF_WATER
            / pvt_collector.absorber.inner_pipe_diameter
        )

    return (
        0.23
        * (current_reynolds_number ** 0.8)
        * (prandtl_number(weather_conditions) ** 0.4)
        * THERMAL_CONDUCTIVITY_OF_WATER
        / pvt_collector.absorber.inner_pipe_diameter
    )


def density_of_water(fluid_temperature: float) -> float:
    """
    The density of water varies as a function of temperature.

    The formula for the density is obtained from:
    https://www.ncbi.nlm.nih.gov/pmc/articles/PMC4909168/

    :param fluid_temperature:
        The temperature of the fluid, measured in Kelvin.

    :return:
        The density of water, measured in kilograms per meter cubed.

    """

    return (
        999.85308
        + 6.32693 * (10 ** (-2)) * (fluid_temperature - ZERO_CELCIUS_OFFSET)
        - 8.523892 * (10 ** (-3)) * (fluid_temperature - ZERO_CELCIUS_OFFSET) ** 2
        + 6.943249 * (10 ** (-5)) * (fluid_temperature - ZERO_CELCIUS_OFFSET) ** 3
        - 3.82126 * (10 ** (-7)) * (fluid_temperature - ZERO_CELCIUS_OFFSET) ** 4
    )


def dynamic_viscosity_of_water(fluid_temperature: float) -> float:
    """
    The dynamic viscosity of water varies as a function of temperature.

    The formula comes from the Vogel-Fulcher-Tammann equation via Wiki:
    https://en.wikipedia.org/wiki/Viscosity#Water

    :param fluid_temperature:
        The temperature of the fluid being modelled, measured in Kelvin.

    :return:
        The dynamic viscosity of water, measured in kilograms per meter second.

    """

    return 0.00002939 * math.exp(507.88 / (fluid_temperature - 149.3))


def free_heat_transfer_coefficient_of_air(
    pvt_collector: pvt.PVT,
    element_top_temperature: float,
    weather_conditions: WeatherConditions,
) -> float:
    """
    Computes the free (conductive/convective) heat-transfer coefficient of air.

    :param pvt_collector:
        The pvt panel being modelled.

    :param element_top_temperature:
        The temperature, measured in Kelvin, of the top-layer of the element.

    :param weather_conditions:
        The weather conditions at the time step being modelled.

    :return:
        The free heat transfer coefficient between the panel and the air, measured in
        Watts per meter squared Kelvin.

    """

    length_scale: float = max(pvt_collector.length, pvt_collector.width)

    current_rayleigh_number = rayleigh_number(
        length_scale,
        element_top_temperature,
        pvt_collector.tilt_in_radians,
        weather_conditions,
    )

    if current_rayleigh_number >= 10 ** 9:
        return (weather_conditions.thermal_conductivity_of_air / length_scale) * (
            0.68
            + (
                (0.67 * current_rayleigh_number ** 0.25)
                / (
                    (1 + (0.492 / prandtl_number(weather_conditions)) ** (9 / 16))
                    ** (4 / 9)
                )
            )
        )

    return (weather_conditions.thermal_conductivity_of_air / length_scale) * (
        nusselt_number(
            length_scale,
            element_top_temperature,
            pvt_collector.tilt_in_radians,
            weather_conditions,
        )
        ** (1 / 4)
    )


def grashof_number(
    length_scale: float,
    surface_temperature: float,
    tilt_in_radians: float,
    weather_conditions: WeatherConditions,
) -> float:
    """
    Computes the dimensionless Grashof number.

    :param length_scale:
        The length scale over which the Grashof number should be computed.

    :param surface_temperature:
        The temperature of the bluff surface for which the Grashof number should be
        computed.

    :param tilt_in_radians:
        The tilt, in radians, between the panel (surface) and the horizontal.

    :param weather_conditions:
        The weather conditions at the time step where the Grashof number should be
        computed.

    :return:
        The dimensionless Grashof number.

    """

    return (
        weather_conditions.thermal_expansivity_of_air  # [1/K]
        * ACCELERATION_DUE_TO_GRAVITY  # [m/s^2]
        * math.cos((pi / 2) - tilt_in_radians)
        * weather_conditions.density_of_air ** 2  # [kg/m^3]
        * length_scale ** 3  # [m^3]
        * abs(surface_temperature - weather_conditions.ambient_temperature)  # [K]
    ) / (
        weather_conditions.dynamic_viscosity_of_air ** 2
    )  # [kg/m*s]


def nusselt_number(
    length_scale: float,
    surface_temperature: float,
    tilt_in_radians: float,
    weather_conditions: WeatherConditions,
) -> float:
    """
    Computes the dimensionless Nusselt number.

    :param length_scale:
        The length scale over which the Grashof number should be computed.

    :param surface_temperature:
        The temperature of the bluff surface for which the Grashof number should be
        computed.

    :param tilt_in_radians:
        The tilt, in radians, between the panel (surface) and the horizontal.

    :param weather_conditions:
        The weather conditions at the time step where the Grashof number should be
        computed.

    :return:
        The dimensionless Nusselt number.

    """

    return 0.825 + (
        0.387
        * rayleigh_number(
            length_scale, surface_temperature, tilt_in_radians, weather_conditions
        )
        ** (1 / 6)
    ) / ((1 + (0.492 / prandtl_number(weather_conditions) ** (9 / 16))) ** (8 / 27))


def prandtl_number(weather_conditions: WeatherConditions) -> float:
    """
    Computes the dimensionless Prandtl number.

    :param weather_conditions:
        The weather conditions at the time step where the Prandtl number should be
        computed.

    :return:
        The dimensionless Prandtl number.

    """

    return (
        weather_conditions.heat_capacity_of_air
        * weather_conditions.dynamic_viscosity_of_air
        / weather_conditions.thermal_conductivity_of_air
    )


def radiative_heat_transfer_coefficient(
    *,
    destination_temperature: float,
    source_emissivity: float,
    source_temperature: float,
    destination_emissivity: Optional[float] = None,
    radiating_to_sky: Optional[bool] = False,
) -> float:
    """
    Computes the radiative heat transfer coefficient between two layers.

    The value computed should always be positive, and any negative flow is computed when
    the net temperature difference is applied to the value returned by this function.

    @@@ BEN-TO-FIX
    The coefficient is computed by using the difference of two squares. This introduces
    an inaccuracy in the model whereby the coefficient is computed at the previous time
    step dispite depending on the temperatures.

    The value for the heat transfer is returned in Watts per meter squared Kelvin.

    :param destination_temperature:
        The temperature of the destination layer/material, measured in Kelvin.

    :param source_temperature:
        The temperature of the source layer/material, measured in Kelvin.

    :param source_emissivity:
        The emissivity of the layer that is radiating, defined between 0 and 1.

    :param destination_emissivity:
        The emissivity of the layer that is receiving the radiation, defined between 0
        and 1. This parameter can be set to `None` in instances where the destination
        has no well-defined emissivity, such as when radiating to the sky.

    :param radiating_to_sky:
        Specifies whether the source of the radiation is the sky (True).

    :return:
        The heat transfer coefficient, in Watts per meter squared Kelvin, between two
        layers, or from a layer to the sky, that takes place by radiative transfer.

    """

    if radiating_to_sky:
        return (
            STEFAN_BOLTZMAN_CONSTANT  # [W/m^2*K^4]
            * source_emissivity
            * (source_temperature ** 2 + destination_temperature ** 2)  # [K^2]
            * (source_temperature + destination_temperature)  # [K]
        )

    if destination_emissivity is None:
        raise ProgrammerJudgementFault(
            "If radiating to an object that is not the sky, a destination emissivity "
            "must be specified."
        )

    # Some cases need to be dealt with where the emissivity of one or more layers is
    # zero. In these cases, the emissivity of the non-zero layer, if there is one, can
    # be used.
    if destination_emissivity == 0 and source_emissivity == 0:
        return 0
    if destination_emissivity == 0:
        return source_emissivity
    if source_emissivity == 0:
        return destination_emissivity

    return (
        STEFAN_BOLTZMAN_CONSTANT  # [W/m^2*K^4]
        * (source_temperature ** 2 + destination_temperature ** 2)  # [K^2]
        * (source_temperature + destination_temperature)  # [K]
    ) / ((1 / source_emissivity) + (1 / destination_emissivity) - 1)


def rayleigh_number(
    length_scale: float,
    surface_temperature: float,
    tilt_in_radians: float,
    weather_conditions: WeatherConditions,
) -> float:
    """
    Computes the non-dimensional Rayleigh number.

    :param length_scale:
        The length scale over which the Grashof number should be computed.

    :param surface_temperature:
        The temperature of the bluff surface for which the Grashof number should be
        computed.

    :param tilt_in_radians:
        The tilt, in radians, between the panel (surface) and the horizontal.

    :param weather_conditions:
        The weather conditions at the time step where the Grashof number should be
        computed.

    :return:
        The dimensionless Grashof number.

    """

    return grashof_number(
        length_scale, surface_temperature, tilt_in_radians, weather_conditions
    ) * prandtl_number(weather_conditions)


def reynolds_number(
    density: float, dynamic_viscosity: float, flow_speed: float, length_scale: float
) -> float:
    """
    Computes the Reynolds number of the flow.

    :param density:
        The density of the fluid, measured in kilograms per meter cubed.

    :param dynamic_viscosity:
        The dynamic viscosity, measured in kilograms per meter second.

    :param flow_speed:
        The speed of the flow, measured in meters per second.

    :param length_scale:
        A characteristic length scale over which Physics in the fluid is occurring.

    :return:
        The dimensionless Reynolds number.

    """

    return (
        density  # [kg/m^3]
        * flow_speed  # [m/s]
        * length_scale  # [m]
        / dynamic_viscosity  # [kg/m*s]
    )


def reduced_temperature(
    ambient_temperature: float, average_temperature: float, solar_irradiance: float
) -> float:
    """
    Computes the reduced temperature of the collector.

    NOTE: The ambient temperature and average temperature need to be measured in the
    same units, whether it's Kelvin or Celcius, but it does not matter which of these
    two is used.

    :param ambient_temperature:
        The ambient temperature surrounding the collector.

    :param average_temperature:
        The average temperature of the collector.

    :param solar_irradiance:
        The solar irradiance, measured in Watts per meter squared.

    :return:
        The reduced temperature of the collector in Kelvin meter squared per Watt.

    """

    return (average_temperature - ambient_temperature) / solar_irradiance


def upward_loss_terms(
    best_guess_temperature_vector: Union[List[float], ndarray],
    pvt_collector: pvt.PVT,
    element: Element,
    source_emissivity: float,
    source_index: int,
    weather_conditions: WeatherConditions,
) -> Tuple[float, float]:
    """
    Computes the upward convective/conductive and radiative loss terms.

    :param best_guess_temperature_vector:
        The best guess at the temperature vector at the current time step.

    :param pvt_collector:
        The pvt panel being modelled.

    :param element:
        The element currently being considered.

    :param source_emissivity:
        The emissivity of the upper layer of the element currently being considered.

    :param source_index:
        The index of the source element.

    :param weather_conditions:
        The weather conditions at the time step being modelled.

    :return:
        A `tuple` containing:
        - the convective/conductive heat transfer upward from the layer,
        - the radiative heat transfer upward from the layer.

    """

    upward_conduction = (
        element.width  # [m]
        * element.length  # [m]
        * _top_heat_transfer_coefficient(
            pvt_collector,
            best_guess_temperature_vector[source_index],
            weather_conditions,
        )  # [W/m^2*K]
    )

    upward_radiation = (
        element.width  # [m]
        * element.length  # [m]
        * radiative_heat_transfer_coefficient(
            destination_temperature=weather_conditions.sky_temperature,
            radiating_to_sky=True,
            source_emissivity=source_emissivity,
            source_temperature=best_guess_temperature_vector[source_index],
        )  # [W/m^2*K]
    )

    return upward_conduction, upward_radiation



#!/usr/bin/python3.7
########################################################################################
# pvt_collector/element.py - Represents a single element within a PVT panel.
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2021
########################################################################################

"""
The element module for the PV-T model.

This module represents a single element within a PV-T panel.

"""

from dataclasses import dataclass
from typing import Optional

__all__ = (
    "Element",
    "ElementCoordinates",
)


@dataclass
class ElementCoordinates:
    """
    Represents the coordinates of the element.

    .. attribute:: x_index
        The x index of the coordinate.

    .. attribute:: y_index
        The y_index of the coordinate.

    """

    x_index: int
    y_index: int

    def __hash__(self) -> int:
        """
        Returns a unique hash based on the two coordinates.

        :return:
            A unique number representing the two coordiantes.

        """

        return int(
            ((self.x_index + self.y_index) * (self.x_index + self.y_index + 1) / 2)
            + self.x_index
        )

    def __str__(self) -> str:
        """
        Return a string representing the coordinates.

        :return:
            A `str` detailing the coordinate information.

        """

        return f"({self.x_index}, {self.y_index})"


@dataclass
class Element:
    """
    Represents a single element within a PV-T panel.

    .. attribute:: absorber
        Whether the absorber layer is present in this element.

    .. attribute:: glass
        Whether the glass layer is present in this element.

    .. attribute:: length
        The length of the element, measured in meters.

    .. attribute:: pipe
        Whether there is a pipe attached to this layer.

    .. attribute:: pv
        Whether the pv layer is present in this element.

    .. attribute:: upper_glass
        Whether an upper-glass layer (i.e., double-glazing) is present.

    .. attribute:: width
        The width of the element, measured in meters.

    .. attribute:: x_index
        The x index for this element.

    .. attribute:: y_index
        The y index for this element.

    .. attribute:: pipe_index
        The index of the attached pipe.

    """

    absorber: bool
    glass: bool
    length: float
    pipe: bool
    pv: bool
    upper_glass: bool
    width: float
    x_index: int
    y_index: int
    pipe_index: Optional[int] = None

    def __str__(self) -> str:
        """
        Return a nice-looking representation of the element.

        :return:
            A `str` giving a nice-looking representation of the element.

        """

        layers = ", ".join(
            [
                entry
                for entry in [
                    "upper-glass" if self.upper_glass else None,
                    "glass" if self.glass else None,
                    "absorber" if self.absorber else None,
                    "pv" if self.pv else None,
                    "pipe" if self.pipe else None,
                ]
                if entry is not None
            ]
        )

        return (
            "Element("
            f"width: {self.width:.3f}m, "
            f"length: {self.length:.3f}m, "
            f"layers: {layers}"
            ")"
        )

    @property
    def coordinates(self) -> ElementCoordinates:
        """
        Returns the coordinates of the element as a element coordinates object.

        :return:
            The element coordinates.

        """

        return ElementCoordinates(self.x_index, self.y_index)



#!/usr/bin/python3.7
########################################################################################
# tank.py - The tank module for the matrix component.
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2021
########################################################################################

"""
The tank module for the matrix component.

This module computes and returns the equation(s) associated with the tank layer of the
PV-T collector for the matrix component.

The model works by arranging the system of differential equations as a matrix equation
such that the temperatures at each time step can be computed based on the coefficients
of the matrix which are computed based on the temperatures of the various components at
the previous time step, as well as various Physical and fundamental constants.

"""


import logging

from typing import List, Optional, Tuple, Union

import numpy

from .. import exchanger, index_handler, tank
from ..pvt_collector import pvt

from ...__utils__ import (
    TemperatureName,
)
from ..__utils__ import WeatherConditions
from ..constants import HEAT_CAPACITY_OF_WATER

__all__ = ("calculate_tank_continuity_equation", "calculate_tank_equation")


def calculate_tank_continuity_equation(
    best_guess_temperature_vector: Union[List[float], numpy.ndarray],
    heat_exchanger: exchanger.Exchanger,
    number_of_temperatures: int,
    pvt_collector: pvt.PVT,
) -> Tuple[List[float], float]:
    """
    Returns a matrix row and resultant vector value representing the tank continuity.

    The HTF flowing through the heat exchanger in the hot-water tank needs to have its
    output temperature computed. The continuity of this fluid is expressed here.

    :param number_of_temperatures:
        The number of temperatures being modelled in the system.

    :return:
        A `tuple` containing:
        - the equation represented as a row in the matrix,
        - and the corresponding value in the resultant method.

    """

    # Compute the row equation
    row_equation: List[float] = [0] * number_of_temperatures

    # If the flow is through the tank heat exchanger:
    if (
        best_guess_temperature_vector[
            index_handler.index_from_temperature_name(
                pvt_collector,
                TemperatureName.tank_in,
            )
        ]
        > best_guess_temperature_vector[
            index_handler.index_from_temperature_name(
                pvt_collector,
                TemperatureName.tank,
            )
        ]
    ):
        row_equation[
            index_handler.index_from_temperature_name(
                pvt_collector,
                TemperatureName.tank,
            )
        ] = (
            -1 * heat_exchanger.efficiency
        )
        row_equation[
            index_handler.index_from_temperature_name(
                pvt_collector,
                TemperatureName.tank_in,
            )
        ] = (
            heat_exchanger.efficiency - 1
        )
        row_equation[
            index_handler.index_from_temperature_name(
                pvt_collector,
                TemperatureName.tank_out,
            )
        ] = 1
        return row_equation, 0

    # Otherwise, the flow is diverted back into the absorber.
    row_equation[
        index_handler.index_from_temperature_name(
            pvt_collector,
            TemperatureName.tank_in,
        )
    ] = -1
    row_equation[
        index_handler.index_from_temperature_name(
            pvt_collector,
            TemperatureName.tank_out,
        )
    ] = 1
    return row_equation, 0


def calculate_tank_equation(
    best_guess_temperature_vector: Union[List[float], numpy.ndarray],
    heat_exchanger: exchanger.Exchanger,
    hot_water_load: float,
    hot_water_tank: tank.Tank,
    logger: logging.Logger,
    number_of_temperatures: int,
    previous_temperature_vector: Optional[numpy.ndarray],
    pvt_collector: pvt.PVT,
    resolution: Optional[int],
    weather_conditions: WeatherConditions,
) -> Tuple[List[float], float]:
    """
    Returns a matrix row and resultant vector value representing the tank equation.

    :param number_of_temperatures:
        The number of temperatures being modelled in the system.

    :return:
        A `tuple` containing:
        - the equation represented as a row in the matrix,
        - and the corresponding value in the resultant method.

    """

    logger.debug("Beginning calculation of Tank equation")

    # Compute the row equation
    row_equation: List[float] = [0] * number_of_temperatures

    tank_internal_energy = (
        hot_water_tank.mass  # [kg]
        * HEAT_CAPACITY_OF_WATER  # [J/kg*K]
        / resolution  # type: ignore  # [s]
    )
    logger.debug("Tank internal energy term: %s W/K", tank_internal_energy)

    hot_water_load_term = hot_water_load * HEAT_CAPACITY_OF_WATER  # [kg/s]  # [J/kg*K]
    logger.debug("Tank hot-water load term: %s W/K", hot_water_load_term)

    heat_loss_term = (
        hot_water_tank.heat_loss_coefficient * hot_water_tank.area  # [W/m^2*K]  # [m^2]
    )
    logger.debug("Tank heat-loss term: %s W/K", heat_loss_term)

    heat_input_term = (
        (
            pvt_collector.absorber.mass_flow_rate  # [kg/s]
            * pvt_collector.absorber.htf_heat_capacity  # [J/kg*K]
            * heat_exchanger.efficiency
        )
        if best_guess_temperature_vector[
            index_handler.index_from_temperature_name(
                pvt_collector,
                TemperatureName.tank_in,
            )
        ]
        > best_guess_temperature_vector[
            index_handler.index_from_temperature_name(
                pvt_collector,
                TemperatureName.tank,
            )
        ]
        else 0
    )
    logger.debug("Tank heat-input term: %s W/K", heat_input_term)

    # Compute the T_t term
    row_equation[
        index_handler.index_from_temperature_name(
            pvt_collector,
            TemperatureName.tank,
        )
    ] = (
        tank_internal_energy + hot_water_load_term + heat_loss_term + heat_input_term
    )

    # Compute the T_c,out term
    row_equation[
        index_handler.index_from_temperature_name(
            pvt_collector,
            TemperatureName.tank_in,
        )
    ] = (
        -1 * heat_input_term
    )

    # Compute the resultant vector value.
    resultant_vector_value = (
        # Internal heat change
        tank_internal_energy  # [W/K]
        * previous_temperature_vector[  # type: ignore
            index_handler.index_from_temperature_name(
                pvt_collector,
                TemperatureName.tank,
            )
        ]  # [K]
        # Hot-water load.
        + hot_water_load_term  # [W/K]
        * weather_conditions.mains_water_temperature  # [K]
        # Heat loss
        + heat_loss_term * weather_conditions.ambient_tank_temperature  # [W/K]  # [K]
    )

    return row_equation, resultant_vector_value



#!/usr/bin/env python
#!/usr/bin/python3
########################################################################################
# heatpanel.py - Primary entry point for the heat-panel package.                       #
#                                                                                      #
# Author: Ben Winchester                                                               #
# Copyright: Ben Winchester, 2021                                                      #
# Date created: 21/09/2021                                                             #
# License: Open source                                                                 #
########################################################################################
"""
heatpanel.py - Primary entry point for the heat-panel package.

"""

import sys

from ..__main__ import main as heat_panel


def main():
    """
    Main function of the CLOVER entry-point script.

    """

    heat_panel(sys.argv[1:])
